;;;-*- Mode:LISP; Package:CHAOS; Base:10; Syntax:Common-lisp -*-
;;; $Id: variable.lisp,v 1.1.1.1 1998/05/16 11:27:43 sawada Exp $
(in-package :chaos)
#|==============================================================================
				 System: Chaos
			       Module: construct
			      File: variable.lisp
==============================================================================|#
#-:chaos-debug
(declaim (optimize (speed 3) (safety 0) #-GCL (debug 0)))
#+:chaos-debug
(declaim (optimize (speed 1) (safety 3) #-GCL (debug 3)))

;;; ********************
;;; VARIABLE DECLARATION _______________________________________________________
;;; ********************

(defun declare-variable-in-module (var-name sort-ref
					    &optional (module *current-module*))
  (let ((mod (if (module-p module)
		 module
		 (find-module-in-env module))))
    (unless mod
      (with-output-panic-message ()
	(princ "internal error, no such module ")
	(print-mod-name module)
	(princ ", variable declaration for ")
	(princ var-name)
	(princ " failed.")
	(chaos-error 'no-such-module)))

    (let ((sort (if (sort-struct-p sort-ref) sort-ref
		    (find-sort-in mod sort-ref))))
      (unless sort
	(with-output-chaos-error ('no-such-sort)
	  (princ "could not find sort ")
	  (princ sort-ref)
	  (princ ", variable declaration for ")
	  (princ var-name)
	  (princ " is ignored.")
	  ))

      ;; check name --
      (when (eql #\` (char (the simple-string (string var-name)) 0))
	(with-output-chaos-error ('invalid-var-decl)
	  (format t "variable name must not start with \"`\",")
	  (print-next)
	  (princ "this is reserved for pseudo variables declared on the fly.")
	  ))
      ;;
      (if (stringp var-name)
	  (setf var-name (intern var-name))
	  (unless (symbolp var-name)
	    (with-output-panic-message ()
	      (princ "internal error: invalid variable name ")
	      (princ var-name)
	      (chaos-error 'invalid-var-decl))))
      
      (let ((old (assoc var-name (module-variables mod)))
	    var)
	(when (and old (sort= sort (variable-sort (cdr old))))
	  (return-from declare-variable-in-module (cdr old)))
	(when old
	  (with-output-chaos-warning ()
	    (princ "variable ")
	    (princ var-name)
	    (princ " once declared as sort ")
	    (princ (string (sort-id (variable-sort (cdr old)))))
	    (princ ", but re-declared as sort ")
	    (princ (string (sort-id sort)))
	    (princ ", ignored.")
	    (return-from declare-variable-in-module nil)
	    ))
	(setf var (make-variable-term sort var-name))
	(push (cons var-name var) (module-variables mod))
	;;
	var))))
 
;;;
;;; DECLARE-ERROR-VARIABLES-IN
;;;
(defun declare-error-variables-in (module)
  (with-in-module (module)
    (dolist (var-decl (module-error-var-decl module))
      (declare-variable var-decl))))

;;; EOF
