;;;-*- Mode:LISP; Package:CAFEIN; Base:10; Syntax:Common-lisp -*-
;;; $Id: match.lisp,v 1.1.1.1 1998/05/16 11:27:44 sawada Exp $
(in-package :chaos)
#|==============================================================================
				  System:Chaos
				 Module:e-match
				File:match.lisp
==============================================================================|#
#-:chaos-debug
(declaim (optimize (speed 3) (safety 0) #-GCL (debug 0)))
#+:chaos-debug
(declaim (optimize (speed 1) (safety 3) #-GCL (debug 3)))

#|
		  EQUATIONAL TERM MATCHING TOP LEVEL ROUTINES
|#
;;; BASED ON matcher of OBJ3 Interpreter of SRI International:
;;; Copyright 1988,1991 SRI International.  All right reserved.

;;; OBSOLETE COMMENTS:
;;; We compute a matcher in a canonicalized order-sorted signature, i.e.,
;;; associated sort of the term is computed in the signature whose sorts are
;;; disjoint, thus ground terms are associated with a single sort, and variables
;;; have a set of disjoint sorts.
;;; 
;;(defmacro match-check-sort (variable-sorts term-sort)
;;  ` (if (sort= (car ,variable-sorts) *universal-sort*)
;;        t
;;        (if (member (car ,term-sort) ,variable-sorts :test #'eq)
;;            t
;;            nil)))
;;

;;; [1]  GENERAL MATCHING 
;;;     TOP-LEVEL ROUTINES _____________________________________________________

;;; FIRST-MATCH : pattern target
;;;                  -> global-state substitution no-match-flag e-equal
;;;-----------------------------------------------------------------------------
;;; returns a first matches of 'pattern' to 'target', that is, all substitutions
;;; \sigma such that \sigma(pattern) = target. (possibly empty) other matchers
;;; will computed by NEXT-MATCH.
;;;
;;; NOTE: Assumptions of the operation:
;;;  (1) target and pattern are flattened;
;;;  (2) target contains no varibles; this guaratees that the range of \sigma is
;;;      disjoint from its domain, i.e., the \sigma is idempotent.
;;;

(defun first-match (t1 t2)
  (declare (type term t1 t2)
	   (values list list (or null t) (or null t)))
  (when *match-debug*
    (format t "~&* First Match --------------------------------~%")
    (princ " t1 = ") (term-print t1)
    (terpri)
    (princ " t2 = ") (term-print t2)
    (force-output))
  (when (term-is-variable? t1) (return-from first-match (values nil nil t nil)))
  (multiple-value-bind (m-sys no-match) 
      (match-decompose&merge (new-match-system t1 t2))
    (when *match-debug*
      (format t "~&result of match-deocmpose&merge, no-match=~a" no-match)
      (force-output))
    ;; Note: if the two terms are similar then "m-sys" is empty.
    (if no-match 
	(values nil nil t nil)		; no match
	(let ((gst (new-global-state)))
	  (declare (type list gst))
	  (cond ((m-system-is-empty? (match-system-sys m-sys))
		 (when *match-debug*
		   (format t "~& return with success"))
		 (values gst 
			 (match-system-to-substitution m-sys) 
			 nil
			 nil))
		((match-system-E-equal m-sys) 
		 (values nil nil nil t))
		(t (multiple-value-bind (sys theory-info)
		       (match-system-extract-one m-sys)
		     (declare (type list sys) (type theory-info theory-info))
		     (when *match-debug*
		       (format t "~& extracted a system ")
		       (print-m-system sys))
		     ;; the matching system is not modified,
		     ;; thus we create a new match-system
		     (multiple-value-bind (th-st no-match)
			 (theory-state-match-initialize theory-info
							sys
							(match-system-env m-sys))
		       (declare (type t th-st) (type (or null t) no-match))
		       (if no-match
			   (values nil nil t nil)
			   (let ((next-gst nil))
			     (when *match-debug*
			       (format t "~&First match calls next-match")
			       (format t "~% old gst: ")
			       (print-global-state gst)
			       )
			     (setq next-gst
				   (global-state-push gst 
						      (match-state-create 
						       (match-system-modif-m-sys
							m-sys sys)
						       sys
						       theory-info
						       th-st)))
			     (when *match-debug*
			       (format t "~& next gst :")
			       (print-global-state next-gst))
			     (multiple-value-bind (new-gst subst no-match)
				 (next-match next-gst)
			       (values new-gst subst no-match nil)))))))
		)))))

;;; NEXT-MATCH : GLOBAL-STATE -> GLOBAL-STATE SUBSTITUTION NO-MATCH-FLAG
;;;-----------------------------------------------------------------------------
;;; next-match suppose that the system on top of gst is fully 
;;; decomposed and merged.

(defun next-match (gst)
  (declare (type list gst)
	   (values list list (or null t)))
  (block the-end 
    (let (st)
      (while (global-state-is-not-empty gst)
	(when *match-debug*
	  (format t "~&* Next-match : global-state = ")
	  (print-global-state gst))
	(setq st (global-state-top gst))
	(multiple-value-bind (new-st no-more)
	    (next-match-state st)
	  (declare (type (or null match-state) new-st)
		   (type (or null t) no-more))
	  (when *match-debug*
	    (format t "~&** Next-match : next-match-state returns no-more = ~a" no-more)
	    (unless no-more
	      (format t "~&-- new state =")
	      (print-match-state new-st)))
	  (if no-more 
	       (setq gst (global-state-pop gst))
	       ;; else
	       (progn
		 (setq gst (global-state-push gst new-st))
		 (let* ((m-sys (match-state-match-system new-st))
			(sys (match-system-sys m-sys)))
		   (when (and (m-system-is-empty? sys)
			      (m-system-is-empty? (match-state-sys-to-solve new-st)))
		     ;; popping: the reasoning is that a successful state
		     ;; also terminates .
		     (setq gst (global-state-pop gst))
		     (when *match-debug*
		       (format t "~&* Next-match : return-with subst"))
		     (return-from the-end
		       (values gst
			       (match-system-to-substitution m-sys)
			       nil))))))
	  )))
    (when *match-debug*
      (format t "~&* Next-match : return with no-match"))
    ;; no match
    (values nil nil t)))

;;; EMPTY-MATCH : PATTERN TARGET -> SUBSTITUTION NO-MATCH-FLAG
;;;-----------------------------------------------------------------------------
;;;
(defun empty-match (t1 t2)
  (declare (type term t1 t2)
	   (values list (or null t)))
  (multiple-value-bind (m-sys no-match) 
      (match-decompose&merge (new-match-system t1 t2))
    (if no-match 
	(values nil t)			; no match
	(cond ((m-system-is-empty? (match-system-sys m-sys))
	       (values (match-system-to-substitution m-sys) nil))
	      (t (values nil t))))))	; no match

;;; MATCHES? : PATTERN TARGET -> BOOL
;;;-----------------------------------------------------------------------------
;;;
(defun matches? (t1 t2)
  (declare (type term t1 t2)
	   (values (or null t)))
  (multiple-value-bind  (gs subst no eeq)
      (first-match t1 t2)
    (declare (ignore gs subst))
    (or eeq (not no))))

;;; FIRST-MATCH-WITH-THEORY : THEORY PATTERN TARGET
;;;                              -> GLOBAL-STATE SUBSTITUTION NO-MATCH E-EQUAL
;;;-----------------------------------------------------------------------------
;;; returns the first matches of 'pattern' to 'target' with fixed operator theory
;;; 'theory'.
;;; 
(defun first-match-with-theory (theory-info t1 t2)
  (declare (type theory-info theory-info)
	   (type term t1 t2)
	   ;; (optimize (debug 3))
	   )
  (multiple-value-bind (m-sys no-match) 
      (match-decompose&merge (new-match-system t1 t2))
    ;; Note that is the two terms are similar then "m-sys" is empty.
    ;; In the current code it is not signaled "E-equal", it must be corrected.
    (if no-match 
	(values nil nil t nil)		; no match
	(let ((gst (new-global-state)))
	  (declare (type list))
	  (cond ((m-system-is-empty? (match-system-sys m-sys))
		 (values gst 
			 (match-system-to-substitution m-sys) 
			 nil
			 nil))
		;; ((match-system-E-equal m-sys) 
		;;   (values nil nil nil t)) ; match & e-equal
		(t (multiple-value-bind (sys th-ign)
		       (match-system-extract-one m-sys)
		     (declare (ignore th-ign))
		     ;; the matching system is not modified,
		     ;; thus we create a new match-system
		     (multiple-value-bind (th-st no-match)
			 (theory-state-match-initialize theory-info
							sys
							(match-system-env m-sys))
		       (if no-match
			   (values nil nil t nil) ; no match
			   (multiple-value-bind (new-gst subst no-match)
			       (next-match (global-state-push 
					    gst 
					    (match-state-create 
					     (match-system-modif-m-sys m-sys sys)
					     sys
					     theory-info
					     th-st)))
			     (values new-gst subst no-match nil nil))))))
		)))))

#|
;;; remainder not used now -------------------------------------------------------
(defun match-decompose&merge-theory (theory m-sys)
  (block no-match
    (let ((sys (match-system-sys m-sys))
	  (env (match-system-env m-sys))
	  (new-env (new-environment) )
	  (new-sys (new-m-system)))
      (dolist (eq (m-system-to-list sys))
	(multiple-value-bind (eq-list clash-of-symbol)
	    (match-decompose-equation-with-theory theory eq)
	  (if clash-of-symbol
	      (return-from no-match (values nil t))
	      (when (match-insert-if-coherent-with 
		     new-env 
		     env 
		     new-sys 
		     eq-list)
		(return-from no-match (values nil t))))))
      (values (create-match-system  new-env
				    new-sys)
	      nil))))

(defun match-decompose-equation-with-theory (theory eq)
  ;; the following curious initialisation allows to modify "list-dec-eq"
  ;; inside match-equation--!decomposition
  (let ((list-dec-eq (cons nil nil))
	(no-match nil))
    (setq no-match (!match-decompose-with-theory theory
						 (equation-t1 eq)
						 (equation-t2 eq)
						 list-dec-eq))
    (values (cdr list-dec-eq) no-match)))


(defun !match-decompose-with-theory (theory t1 t2 l-result)
  (block no-match
    (cond
      ((term-is-variable? t1)
       ;; test of the sort of the variable,
       (when (sort<= (sort-of t2) (variable-sort t1) *current-sort-order*)
	 (push (make-equation t1 t2) (cdr l-result))
	 (return-from !match-decompose-with-theory nil))
       (!match-decompose-on-demand-with-theory theory t1 t2 l-result))
      ((term-is-variable? t2)
       (return-from no-match 'no-match-to-var))

      ((term-is-builtin-constant? t1)
       (if (term-builtin-equal t1 t2)
	   nil
	   (return-from no-match t)))
      ((term-is-builtin-constant? t2) (return-from no-match t))
      (t (let ((t1-top (term-method t1)))
	   (cond
	     ((not (theory-is-empty-for-matching theory))
	      (push (make-equation t1 t2) (cdr l-result))
	      (return-from no-match nil))
	     (t (let ((t2-top (term-method t2)))
		  (if (method-is-of-same-operator t1-top t2-top)
		      (let ((t1-subterms (term-subterms t1))
			    (t2-subterms (term-subterms t2)))
			(do ((tt1)
			     (tt2))
			    ((null t1-subterms))
			  (setq tt1 (pop t1-subterms) 
				tt2 (pop t2-subterms))
			  (let ((not-ok (!match-decompose-with-theory theory
								      tt1
								      tt2
								      l-result)))
			    (when (and not-ok
				       (!match-decompose-on-demand-with-theory theory
									       t1
									       t2
									       l-result))
			      (return-from no-match t))))
			(return-from no-match nil))
		      (!match-decompose-on-demand-with-theory theory
							      t1 t2 l-result)
		      ))))
	   )))))

(defun !match-equation-decompose-on-demand-with-theory (theory t1 t2 l-result)
  (if (term-is-on-demand? t2)
      (progn
	(mark-term-as-lowest-parsed t2)
	(if (normalize-term t2)
	    t
	    (!match-decompose-with-theory theory t1 t2 l-result)))
      t))
|#


