;;;-*- Mode: Lisp; Syntax:CommonLisp; Package:CHAOS; Base:10 -*-
;;; $Id: gen-print.lisp,v 1.5 1998/07/06 08:14:42 sawada Exp $
(in-package :chaos)
#|=============================================================================
				  System:CHAOS
			       Module: primitives
			      File: gen-print.lisp
=============================================================================|#
#-:chaos-debug
(declaim (optimize (speed 3) (safety 0) #-GCL (debug 0)))
#+:chaos-debug
(declaim (optimize (speed 1) (safety 3) #-GCL (debug 3)))

;;; *************
;;; TERM PRINTERS______________________________________________________________
;;; *************

;;; TERM-PRINT : TERM STREAM -> VOID
;;; print term TERM to stream STREM.
;;; the printer controll global *fancy-print* is non-nil, the term will be
;;; printed in a fancy manner.

;;; **TODO** for coded terms

(declaim (special *current-term-depth*)
	 (type fixnum *current-term-depth*))
(defvar *current-term-depth* 0)

(defun term-print (term &optional (stream *standard-output*) (print-var-sort t))
  (declare (type (or null term) term)
	   (type stream stream)
	   (type (or null t) print-var-sort)
	   (values t))
  (if *fancy-print*
      (term-print2 term parser-max-precedence stream print-var-sort)
      (term-print1 term stream print-var-sort))
  )

;;; VARIABLE

(defun print-variable (term &optional (stream *standard-output*))
  (let ((*standard-output* stream)
	(body (term-body term)))
    (princ (string (variable$name body)))
    (princ ":")
    (princ (sort-id (variable$sort body)))))

;;; AST ::= "(" keyword-symbol . arguments ")"
;;;

(defun print-ast-vd (ast stream print-var-sort)
  (print-check)
  (cond ((consp ast)
	 (let ((flg nil))
	   (format stream ":~a(" (car ast))
	   (dolist (arg (cdr ast))
	     (if flg (princ ","))
	     (term-print arg stream print-var-sort)
	     (setf flg t))
	   (princ ")")))
	(t (princ ast))))

(defun term-print1 (term &optional (stream *standard-output*) (print-var-sort t))
  (declare (type (or null term) term)
	   (type stream stream)
	   (type (or null t) print-var-sort)
	   (values t))
  (let ((*standard-output* stream)
	(body (term-body term))
	(*current-term-depth* (1+ *current-term-depth*)))
    ;;
    (when (and *term-print-depth*
	       (> *current-term-depth*
		  (the fixnum *term-print-depth*)))
      (princ " ... ")
      (return-from term-print1 nil))
    (unless term
      (princ "NIL!")			; this should not happen in general,
					; but may happen while debug output.
      (return-from term-print1 nil))
    (print-check)
    (cond ((term$is-variable? body)
	   (princ (if (numberp (variable$name body))
		      (variable$name body)
		      (string (variable$name body))))
	   (when print-var-sort
	     (when (and (let ((sort (variable$sort body)))
			  (not (or (eq sort *universal-sort*)
				   (eq sort *huniversal-sort*)
				   (eq sort *bottom-sort*)
				   (eq sort *cosmos*))))
			(or *print-with-sort*
			    (not (assoc (variable$name body)
					(module-variables *current-module*)))))
	       (princ ":")
	       (print-sort-name (variable$sort body) *current-module*))))
	  ((term$is-psuedo-constant? body)
	   (let ((val (term$builtin-value body)))
	     (format t "~a:" (if (numberp val)
				 val
				 (string val)))
	     (princ (string (sort-id (term$sort body))))))
	  ((term$is-builtin-constant? body)
	   (let ((val (term$builtin-value body)))
	     (cond ((stringp val)
		    (if (not (sort= (term$sort body)
				    *sort-id-sort*))
			(progn (princ "\"") (princ val) (princ "\""))
			(princ val)))
		   ((symbolp val)	; assume this is a psude variable?
		    (princ (string val)))
		   ((and (bsort-p (term$sort body))
			 (bsort-term-printer (term$sort body)))
		    (funcall (bsort-term-printer (term$sort body))
			     val))
		   ((characterp val)
		    (print-character val))
		   (t (prin1 val)))))
	  ((term$is-lisp-form? body)
	   (if (term$is-simple-lisp-form? body)
	       (princ "#! ")
	       (princ "#!! "))
	   (princ (term$lisp-form-original-form body)))
	  ((term$is-applform? body)
	   (let* ((hd (term$head body))
		  (op nil))
	     (setq op (method-operator hd))
	     (cond ((not (operator-is-mixfix op))
		    ;; patch. 
		    ;; (princ (car (operator-symbol op)))
		    (format t "~{~a~^ ~}" (operator-symbol op))
		    (let ((subs (term$subterms body)))
		      (when subs
			(princ "(")
			(let ((flg nil))
			  (dolist (i subs)
			    (if flg
				(progn (princ ",")
				       (print-check))
				(setq flg t))
			    (term-print1 i stream print-var-sort)))
			(princ ")")
			(print-check))))
		   (t (princ "(")
		      (let ((subs (term$subterms body))
			    (prv nil))
			(dolist (i (operator-token-sequence op))
			  (when prv
			    (princ " ")
			    (print-check))
			  (setq prv i)
			  (cond ((eq t i)
				 (term-print1 (car subs) stream print-var-sort)
				 (setq subs (cdr subs)))
				(t (princ i))))
			(princ ")"))))))
	  (t (format stream "~s" body))	; what is this?
	  )))

(defun term-print2 (term prec
			 &optional (stream *standard-output*) (print-var-sort t))
  (declare (type (or null term) term)
	   (type fixnum prec)
	   (type stream stream)
	   (type (or null t) print-var-sort)
	   (values t))
  (let ((*standard-output* stream)
	(*current-term-depth* (1+ *current-term-depth*)))
    ;;
    (when (and *term-print-depth*
	       (> *current-term-depth*
		  (the fixnum *term-print-depth*)))
      (princ " ... ")
      (return-from term-print2 nil))
    ;;
    (print-check)
    (unless term
      (princ "NIL!")			; this should not happen in general,
					; but may happen while debug output.
      (return-from term-print2 nil))
    (cond ((term-is-variable? term)
	   (princ (if (numberp (variable-name term))
		      (variable-name term)
		      (string (variable-name term))))
	   (when print-var-sort
	     (when (and (let ((sort (variable-sort term)))
			  (not (or (eq sort *universal-sort*)
				   (eq sort *huniversal-sort*)
				   (eq sort *bottom-sort*)
				   (eq sort *cosmos*))))
			(or *print-with-sort*
			    (not (assoc (variable-name term)
					(module-variables *current-module*)))))
	       (princ ":")(print-sort-name (variable-sort term)
					   *current-module*))))
	  ((term-is-psuedo-constant? term)
	   (let ((val (term-builtin-value term)))
	     (format t "~a:" (if (numberp val)
				 val
				 (string val)))
	     (princ (string (sort-id (term-sort term))))))
	  ((term-is-builtin-constant? term)
	   (let ((val (term-builtin-value term)))
	     (cond ((stringp val)
		    (if (not (sort= (term-sort term)
				    *sort-id-sort*))
			(progn (princ "\"") (princ val) (princ "\""))
			(princ val)))
		   ((symbolp val)
		    (princ (string val)))
		   ((and (bsort-p (term-sort term))
			 (bsort-term-printer (term-sort term)))
		    (funcall (bsort-term-printer (term-sort term))
			     val))
		   ((characterp val)
		    (print-character val))
		   (t (princ val)))))
	  ((term-is-lisp-form? term)
	   (if (term-is-simple-lisp-form? term)
	       (princ "#! ")
	       (princ "#!! "))
	   (princ (lisp-form-original-form term)))
	  ((term-is-applform? term)
	   (let* ((hd (term-head term))
		  (op nil))
	     (setq op (method-operator hd))
	     (cond ((not (operator-is-mixfix op))
		    (format t "~{~a~^ ~}" (operator-symbol op))
		    (let ((subs (term-subterms term)))
		      (when subs
			(princ "(")
			(let ((flg nil))
			  (dolist (i subs)
			    (if flg (progn (princ ",")(print-check))
				(setq flg t))
			    (term-print2 i parser-max-precedence
					 stream print-var-sort)
			    ))
			(princ ")")
			(print-check))))
		   (t (let ((prec-test (and (get-method-precedence hd)
					    (<= prec (get-method-precedence hd))))
			    (assoc-test (method-is-associative hd)))
			(when prec-test (princ "("))
			(let ((subs (term-subterms term))
			      (prv nil))
			  (dolist (i (operator-token-sequence (method-operator hd)))
			    (cond
			      ((eq i t)
			       (when prv (princ " "))
			       (setq prv t)
			       (let ((tm (car subs)))
				 (term-print2 tm
					      (if (and assoc-test
						       tm
						       (term-is-application-form? tm)
						       (method-is-of-same-operator
							(term-head term) (term-head tm)))
						  parser-max-precedence
						  (or (get-method-precedence hd) 0))
					      stream
					      print-var-sort)
				 (setq subs (cdr subs))))
			      (t (when prv (princ " ")(print-check))
				 (setq prv t)
				 (princ i)(print-check)))))
			(when prec-test (princ ")")(print-check))
			)))))
	  (t (format stream "~s" (term-body term)))
	  )))

;;;-----------------------------------------------------------------------------
;;; PRINT-TERM-SEQ : List[TERM]
;;; print sequence of terms separating each ",".
;;;
(defun print-term-seq (x)
  (declare (type list x)
	   (values t))
  (if (null x)
      (princ "NULL")
      (let ((flg nil))
	(dotimes (i (length x))
	  (declare (type fixnum i))
	  (if flg
	      (princ ",")
	      (setq flg t))
	  (princ " ")
	  (let ((e (elt x i)))
	    (term-print e)))
	)))

;;; PRINT-TERM-TREE
(defvar *show-sort* nil)

(defun print-term-tree (tree &optional (show-sort nil) (stream *standard-output*))
  (!print-term-tree tree show-sort stream nil))

(defun print-term-graph (tree &optional (show-sort nil) (stream *standard-output*))
  (!print-term-tree tree show-sort stream t))

(defun !print-term-tree (tree show-sort stream show-as-graph)
  (let* ((*show-sort* show-sort)
	 (leaf?
	  #'(lambda (tree) (or (term-is-variable? tree)
			       (term-is-constant? tree))))
	 (leaf-name
	  #'(lambda (term)
	      (let ((name (if (term-is-builtin-constant? term)
			      (let ((value (term-builtin-value term)))
				(cond ((characterp value)
				       (format nil "'~a'" value))
				      ((symbolp value)
				       (string value))
				      (t (format nil "~s" value))))
			      (if (term-is-variable? term)
				  (string (variable-name term))
				  (if (term-is-lisp-form? term)
				      (lisp-form-original-form term)
				      (format nil "~{~a~}"
					      (method-symbol (term-head term)))))))
		    (sort (term-sort term)))
		(if *show-sort*
		    (format nil "~a:~a" name (string (if sort
							 (sort-id sort)
							 "unknown")))
		    name))))
	 (leaf-info
	  #'(lambda (tree) (declare (ignore tree)) t))
	 (int-node-name
	  #'(lambda (tree) (funcall leaf-name tree)))
	 (int-node-children 
	  #'(lambda (tree) (term-subterms tree))))
    (force-output stream)
    (print-next *print-indent* stream)
    (print-trees (list (if show-as-graph
			   (augment-tree-as-graph tree)
			   (augment-tree tree)))
		 stream)))

;;; ***************
;;; GENERIC PRINTER_____________________________________________________________
;;; ***************

;;; PRINT-AST : ast [ readable? output-stream ] -> void
;;;
;;; print out Chaos AST form to `output-stream'. the stream is defaulted to 
;;; *standard-output*.
;;; if `readable?' is T, the printed form can be evaluated by `eval-ast',
;;; otherwise, more descriptive output will be generated(NOT YET).
;;;

(defun print-ast (ast &optional (readable? t) (stream *standard-output*))
  (declare (ignore readable?))
  (print-chaos-object ast stream))

;;; PRINT-CHAOS-OBJECT
;;; simple printer for AST, TERM, INTERNAL OBJECTS.
;;;
(defun print-chaos-object (object &optional (stream *standard-output*))
  (cond ((chaos-ast? object)
	 (let ((printer (ast-printer object)))
	   (if printer
	       (let ((mod (or *current-module* *last-module*)))
		 (if mod
		     (with-in-module (mod)
		       (funcall printer object stream))
		     (funcall printer object stream)))
	       (let ((*print-circle* nil)
		     (*print-pretty* t))
		 (prin1 object stream)))))
	((and (chaos-object? object) (not (stringp object)))
	 (let ((printer (object-printer object)))
	   (if printer
	       (let ((mod (or *current-module*
			      *last-module*)))
		 (if mod
		     (with-in-module (mod)
		       (funcall printer object stream))
		     (funcall printer object stream)))
	       (let ((*print-circle* nil)
		     (*print-pretty* t))
		 (prin1 object stream)))))
	((term? object)
	 (let ((mod (or *current-module* *last-module*)))
	   (if mod
	       (with-in-module (mod)
		 (term-print object stream))
	       (term-print object stream))))
	((opinfo-p object) 
	 (fresh-line stream)
	 (print-chaos-object (opinfo-operator object) stream)
	 (format stream "~&-- delcrations -------------------")
	 (dolist (meth (opinfo-methods object))
	   (print-next)
	   (print-chaos-object meth stream)))
	((is-operator-theory? object)
	 (print-theory-brief object stream))
	((consp object)
	 (princ "(" stream)
	 (let ((*print-indent* (+ 2 *print-indent*))
	       (flg nil))
	   (do ((elem object (cdr elem)))
	       ((not (consp elem))
		(unless (listp elem)
		  (princ " . " stream) (print-chaos-object elem stream)))
	     (when flg
	       (princ " " stream))
	     (print-chaos-object (car elem)  stream)
	     (setq flg t)))
	 (princ ")" stream))
	(t (let ((*print-circle* nil)
		 (*print-pretty* t))
	     (prin1 object stream)))))
	   
(defun print-obj-list (lst)
  (let ((flag nil))
    (do ((l lst (cdr l)))
	((atom l) (unless (null l) (print-chaos-object l nil)))
      (if flag (print-next) (setq flag t))
      (print-chaos-object (car l)))))

;;; SOME SPECIFIC PRINTERS
;;;-----------------------------------------------------------------------------
(defun print-theory-brief (thy &optional (stream *standard-output*))
  (let ((th-info (theory-info thy))
	(flag nil)
	(val (theory-zero thy)))
    ;; (when (theory-info-is-empty-for-matching th-info)
    ;; (princ "empty")
    ;; (return-from print-theory-brief nil))
    (when (or (theory-info-is-AC th-info)
	      (theory-info-is-A th-info)
	      (theory-info-is-AI th-info)
	      (theory-info-is-AZ th-info)
	      (theory-info-is-AIZ th-info)
	      (theory-info-is-ACI th-info)
	      (theory-info-is-ACZ th-info)
	      (theory-info-is-ACIZ th-info))
      (setq flag t)
      (princ "assoc" stream))
    (when (or (theory-info-is-AC th-info)
	      (theory-info-is-C th-info)
	      (theory-info-is-CI th-info)
	      (theory-info-is-CZ th-info)
	      (theory-info-is-CIZ th-info)
	      (theory-info-is-ACI th-info)
	      (theory-info-is-ACZ th-info)
	      (theory-info-is-ACIZ th-info))
      (if flag (princ " " stream) (setq flag t))
      (princ "comm" stream))
    (when (or (theory-info-is-I th-info)
	      (theory-info-is-IZ th-info)
	      (theory-info-is-CI th-info)
	      (theory-info-is-CIZ th-info)
	      (theory-info-is-AI th-info)
	      (theory-info-is-AIZ th-info)
	      (theory-info-is-ACI th-info)
	      (theory-info-is-ACIZ th-info))
      (if flag (princ " " stream) (setq flag t))
      (princ "idem" stream))
    (when val
      (if flag (princ " " stream))
      (if (null (cdr val)) (princ "id: " stream) (princ "idr: " stream))
      (term-print (car val) stream))
    ))

(defun print-theory (th)
  (print-theory-info (theory-info th))
  (princ " zero: ")
  (let* ((zs (theory-zero th))
	 (flag (and (consp zs)
		    (eq 'to-rename (car zs)))))
    (when flag
      (princ "to rename: ")
      (setq zs (cadr zs)))
    (if zs
	(progn
	  (term-print (car zs))
	  (when (cdr zs) (princ " rule-only")))
	(princ "NONE"))
    ))

(defun print-theory-info (thinf)
  (prin1 (theory-info-name thinf))
  (princ "[") (prin1 (theory-info-code thinf)) (princ ",")
  (print-check)
  (unless (theory-info-is-empty-for-matching thinf) (princ "not "))
  (princ "empty for matching,equal:")
  (prin1 (theory-info-match-equal-fun thinf))
  (princ ",")
  (print-next)
  (princ "init:")
  (prin1 (theory-info-match-init-fun thinf))
  (princ ",")
  (print-check)
  (princ "next:")
  (prin1 (theory-info-match-next-fun thinf))
  (princ "]"))

;;; little helper
(defun print-simple-princ-flat (x)
  #||
  (when (and (fboundp 'filecol)
	     (< *print-line-limit* (filecol *standard-output*)))
    (print-next))
  ||#
  (cond ((null x))
	((stringp x) (princ x))
	((consp x)
	 (let ((flag nil)
		 (tail x))
	     (loop
	      (when (not (consp tail)) (return))
	      (if flag
		  (unless (or (null (car tail))
			      (equal "," (car tail)))
	     (princ " "))
		  (setq flag t))
	      (print-simple-princ-flat (car tail))
	      (setq tail (cdr tail)))
	     (when tail (princ " ... ") (prin1 tail))
	     ))
	(t (print-chaos-object x))))

;;; HASH TABLE

(defun dump-chaos-hash (hash-table &optional (title "chaos hash table dump"))
  (format t "~&~a__________________________" title)
  (maphash #'(lambda (key value)
	       (format t "~&key(type:~a) = | ~a ~%  " (type-of key) key)
	       (print-chaos-object key)
	       (format t "~&value(type:~a, ~a) = " (type-of value)
		       (if (chaos-object? value)
			   (object-type value)
			   (if (chaos-ast? value)
			       (ast-type value)
			       'unknown)))
	       (print-chaos-object value)
	       (format t "~&----------------------------------------"))
	   hash-table))

;;; ASSOC TABLE
(defun dump-chaos-assoc-table (table &optional (title "chaos assoc table dump"))
  (format t "~&~a__________________________" title)
  (dolist (entry table)
    (let ((key (car entry))
	  (value (cdr entry)))
      (format t "~&key(type:~a) = " (type-of key))
      (print-chaos-object key)
      (format t "~&value(type:~a, ~a) = "
	      (type-of value)
	      (if (chaos-object? value)
		  (object-type value)
		  (if (chaos-ast? value)
		      (ast-type value)
		      (type-of value)
		      )))
      (print-chaos-object value)
      (format t "~&----------------------------------------"))))
  
(defun dump-modexp-local ()
  (dump-chaos-assoc-table *modexp-local-table* "Moduexp Local "))
    
(defun dump-modules-so-far ()
  (dump-chaos-assoc-table *modules-so-far-table* "Module so far "))

(defun dump-modexp-eval ()
  (dump-chaos-assoc-table *modexp-eval-table* "Modexp eval table "))

(defun dump-modexp-view ()
  (dump-chaos-assoc-table *modexp-view-table* "Modexp view "))

;;; EOF
