;;;-*- Mode:Lisp; Syntax:Common-Lisp; Package:CHAOS -*-
;;; $Id: dag.lisp,v 1.1.1.1 1998/05/16 11:27:46 sawada Exp $
(in-package :CHAOS)
#|==============================================================================
				 System: Chaos
				 Module: comlib
				 File: dag.lisp
==============================================================================|#
#-:chaos-debug
(declaim (optimize (speed 3) (safety 0) #-GCL (debug 0)))
#+:chaos-debug
(declaim (optimize (speed 1) (safety 3) #-GCL (debug 3)))

;;; dag node = ( datum List[node] . flag )
;;; flag is used for dag traverse.

(defstruct dag-node
  (datum nil :type t)
  (subnodes nil :type list)
  (flag nil :type (or null t)))

(defmacro create-dag-node (datum subnodes)
  `(make-dag-node :datum ,datum :subnodes ,subnodes))

(defmacro dag-node-is-marked? (node) `(dag-node-flag ,node))
(defmacro mark-dag-node (node) `(setf (dag-node-flag ,node) t))
(defmacro unmark-dag-node (node) `(setf (dag-node-flag ,node) nil))

(defun unmark-all-dag-nodes (dag)
  (declare (type dag-node dag))
  (unmark-dag-node dag)
  (dolist (sub (dag-node-subnodes dag))
    (declare (type dag-node sub))
    (unmark-all-dag-nodes sub)))

;;; dag-dfs : dag function
;;; traversing dag depth first manner, apply function to each node.
;;;
(defun dag-dfs (dag &optional (function #'identity))
  (declare (type dag-node dag)
	   (values t))
  (unmark-all-dag-nodes dag)
  (do-dag-dfs dag function))

(defun do-dag-dfs (dag function)
  (declare (type dag-node dag)
	   (type (or symbol function) function)
	   (values t))
  (unless (dag-node-is-marked? dag)
    (mark-dag-node dag)
    (dolist (sub (dag-node-subnodes dag))
      (unless (dag-node-is-marked? sub)
	(do-dag-dfs sub function)))
    (funcall function dag)))

;;; EOF
