;;;-*- Mode:LISP; Package:CHAOS; Base:10; Syntax:Common-lisp -*-
;;; $Id: globals.lisp,v 1.14 1999/03/03 14:43:26 sawada Exp $
(in-package :chaos)
#|==============================================================================
				 System: CHAOS
				 Module: comlib
			       File: globals.lisp
==============================================================================|#
#-:chaos-debug
(declaim (optimize (speed 3) (safety 0) #-GCL (debug 0)))
#+:chaos-debug
(declaim (optimize (speed 1) (safety 3) #-GCL (debug 3)))

;;; == DESCRIPTION =============================================================
;;; This file gathers the declaration of global/special variables of CHAOS
;;; system.
;;;

;;;********************
;;; RUNTIME ENVIRONMENT_________________________________________________________
;;;********************
;;;
;;; In Chaos, all of the works are done in a `context'.
;;; A context is a dag of modules with importation relation as arrows,
;;; the top node (module) is called `current module'.
;;; 
;;; The current module and its some frequently accessed informations are
;;; bound to global variables providing run time environment.
;;;
;;;   *current-module*       : the current-module
;;;   *current-sort-order*   : transitive closure of sort relations of
;;;                            current-mdoule.
;;;   *current-opinfo-table* : operator information table of the current-module.
;;;   *current-ext-rule-table* : extented rules for A, and AC
;;;

(declaim (special *current-module*	; the module currently the target of
					; operations.
		  *current-sort-order*	; closure of sort relations of current
					; module.
		  *current-opinfo-table* ; operator information table of the
					; current module.
		  *current-ext-rule-table*
		  *top-level-definition-in-progress*
		  ))

(defvar *top-level-definition-in-progress* nil)

;;; *last-module* bounds the module object which was the target of the operation
;;; in the last time, i.e., whenever the context is swithced, *last-module*
;;; bounds the last `current module'.
;;; *open-module* bounds the 'opening' module.
;;; 
(declaim (special *open-module*		; the module crrently opened.
		  *last-module*		; the module which was the current last
					; time. 
		  *last-before-open*	; the module which was *last* before the
					; currently opened module.
		  ))

(defvar *current-module* nil)
(defvar *current-sort-order* nil)
(defvar *current-opinfo-table* nil)
(defvar *current-ext-rule-table* nil)
(defvar *open-module* nil)
(defvar *last-module* nil)
(defvar *last-before-open* nil)

;;; Feature for require & provide
;;;
(defvar *chaos-features* nil)

;;; Some top level flags & variables.--------------------------------------------

(declaim (special *chaos-verbose* *chaos-quiet* *chaos-input-source*))
(defvar *chaos-verbose* nil)
(defvar *chaos-quiet* nil)
(defvar *chaos-input-source* nil)	; binds a file name when processing
					; input from the file. 
(declaim (special *chaos-input-level*)
	 (type (integer 0 11)
	       *chaos-input-level*
	       *chaos-input-nesting-limit*))

(defvar *chaos-input-level* 0)
(defvar *chaos-input-nesting-limit* 10)

(declaim (special *auto-context-change*))
(defvar *auto-context-change* nil)

(defvar *system-prelude-dir* nil)
(defvar *system-lib-dir* nil)
(defvar *system-ex-dir* nil)

;;; ***********************************
;;; CONTROLL FLAGS OF REWRITING PROCESS_________________________________________
;;; ***********************************

(declaim (special *trace-level*))
(declaim (type (integer 0 256) *trace-level*))
(defvar *trace-level* 0)
(declaim (special *self*))
(defvar *self* nil)
(defvar $$cond nil)
(defvar $$trace-rewrite nil)		; flag, non-nil -> trace.
(defvar $$trace-rewrite-whole nil)	; flag, non-nil -> trace whole.
(defvar *rewrite-stepping* nil)		; flag, non-nil -> under stepping.
(declaim (type (or null fixnum) *rewrite-count-limit*))
(defvar *rewrite-count-limit* nil)
					; flag, non-nil(integer) -> limitation
					; for rewriting steps.
(defvar *rewrite-stop-pattern* nil)	; flag, non-nil(term) -> stop rewriting
					; iff matches to the pattern.
(defvar *steps-to-be-done* nil)		; remaining steps before stop
(defvar $$mod 'void)
;;;
(defvar *old-context* nil)
(declaim (special *old-context*))
(declaim (special *allow-$$term*))
(defvar *allow-$$term* t)
(defvar $$term 'void)			; current target term, destructively
					; modified
(defvar $$subterm nil)			; subterm of $$term selected
(defvar $$term-context nil)		; context module of $$term
(defvar $$selection-stack nil)
(defvar $$action-stack nil)
;;;
(defvar $$norm 'void)
(defvar $$show-red nil)
(defvar *perform-on-demand-reduction* t)
(declaim (special *rewrite-exec-mode*))
(defvar *rewrite-exec-mode* nil)
(declaim (special *cexec-target*))
(defvar *cexec-target* nil)
(declaim (special *rewrite-exec-condition*))
(defvar *rewrite-exec-condition* nil)
(declaim (special *rewrite-semantic-reduce*))
(defvar *rewrite-semantic-reduce* nil)
(declaim (type fixnum *rule-count*))
(defvar *rule-count* 0)
(defvar *show-stats* t)
(defvar *try-try* nil)
(defvar *reduce-conditions* nil)
(declaim (type fixnum $$trials))
(defvar $$trials 1)
(declaim (type fixnum $$matches))
(defvar $$matches 0)
(defvar *on-reduction* t)
(defvar *reduce-builtin-eager* nil)
(declaim (type fixnum *condition-trial-limit*))
(defparameter .condition-trial-limit-default.
  #+GCL 240
  #-GCL 5500)
(defvar *condition-trial-limit* .condition-trial-limit-default.)

;;; MEL support
(defvar *mel-sort* nil)
(defvar *mel-always* nil)
;;; (defvar *mel-reduce* nil)

;; memoization
(defvar *memo-rewrite* t)		; use memo mechanism
(defvar *clean-memo-in-normalize* nil)
(declaim (special *hash-hit*)
	 (type (unsigned-byte 29) *hash-hit*))
(defvar *term-memo-hash-hit* 0)

(defvar *allow-illegal-beh-axiom* t)

;;;*********************
;;; REGULARIZING CONTROL_______________________________________________________
;;;*********************
(declaim (special *regularize-signature*))
(defvar *regularize-signature* t)

;;;***************************
;;; Flags for printer control.__________________________________________________
;;;***************************

(declaim (special *print-indent*))
(declaim (type (integer 0 128)
	       *print-line-limit* *chaos-print-level*
	       *print-indent* *print-indent-increment*
	       *print-length*))
;; (defvar *chaos-verbose* nil)
(defvar *module-all-rules-every* nil)
(defvar *fancy-print* t)
(defvar *print-indent* 0)
(defparameter *print-indent-increment* 1)
(defvar *print-explicit* nil)		;if t then give more detail on sorts, etc.
(defvar *print-abbrev-mod* nil)		; abbreviate module names
(defvar *print-abbrev-num* 0)
(defvar *print-abbrev-table* nil)
(defvar *print-abbrev-quals* nil)
(defvar *print-with-sort* nil)
(defvar *print-operator-table* nil)
(defvar *print-flag-module-values* nil)
(defvar *print-indent-contin* nil)
(defvar *print-line-limit* 64)
(defvar *print-mode* nil)
(defvar *print-all-eqns* nil)
(defvar *print-ignore-mods* nil)
(defvar *chaos-print-level* 5)
(defvar *chaos-print-length* 100)
(defvar *chaos-print-errors* nil)
(defvar *chaos-input-quiet* nil)
(defvar *print-variables* nil)
;;;
(declaim (type (or null fixnum) *term-print-depth*))
(defvar *term-print-depth* nil)

;;; NOT USED
;;; GRAMMAR
;;; 
;;; *GRM.CURRENT* is a grammar to which the following operations are to be applied. 
;;;         define.grm-rule
;;;         grm.clear-rule
;;; set by 'in.grammar' operation.
;;;
;;;(defvar *grm.current* nil)

;;; *CHAOS.CURRENT-GRAMMAR* is a list of grammar with which the `grm.parse' works.
;;; set by 'use.grammar' operation. (see "term-parser.lisp")
;;;
;;; (defvar *chaos.current-grammar* nil)

;;; ERROR TAGS ----------------------------------------------------------------
;;; (defvar *module-not-found-error* 'Module-Not-found)
;;; (defvar *sort-not-found-error* 'sort-not-found)
;;; (defvar *operator-not-found-error* 'operator-not-found)
;;;

;;;=============================================================================
;;; BUILTIN SORTS
;;;=============================================================================
;; NAMES of Builtin sort.
(defconstant $name-sort '|_ Sort _|)
(defconstant $name-gen-sort '|_ GeneralSort _|)
(defconstant $name-bi-sort '|_ BuiltinSort _|)
(defconstant $name-identifier '|Identifier|)
(defconstant $name-cosmos '|_ Cosmos _|)
(defconstant $name-universal '|_ Universal _|)
(defconstant $name-huniversal '|_ HUniversal _|)
(defconstant $name-bottom '|_ Bottom _|)
(defconstant $name-hbottom '|_ HBottom _|)
(defconstant $name-record '|_ Record _|)
(defconstant $name-class '|_ Class _|)
(defconstant $name-and-sort '|_ ^-Sort _|)
(defconstant $name-or-sort '|_ \|-Sort _|)
(defconstant $name-err-sort '|_ ?-Sort _|)
(defconstant $name-operator '|_ Operator _|)
(defconstant $name-module '|_ Module _|)
(defconstant $name-axiom '|_ Axiom _|)

;;; builtin sorts
(defvar *cosmos* 'void)			; the whole 
(defvar *universal-sort* 'void)		; visible universe
(defvar *huniversal-sort* 'void)	; hidden universe
(defvar *bottom-sort* 'void)		; visible bottom sort
(defvar *hbottom-sort* 'void)		; hidden bottom sort
(defvar *sort-sort* 'void)
(defvar *general-sort* 'void)
(defvar *builtin-sort* 'void)
(defvar *identifier-sort* 'void)
(defvar *qid-sort* 'void)
(defvar *syntax-err-sort* 'void)
(defvar *type-err-sort* 'void)
(defvar *op-err-sort* 'void)
(defvar *and-sort* 'void)
(defvar *or-sort* 'void)
(defvar *err-sort* 'void)
(defvar *sort-error* 'none)
(defvar *record-sort* 'void)
(defvar *class-sort* 'void)
(defvar *operator-sort* 'void)
(defvar *module-sort* 'void)
(defvar *bool-sort* 'void)
(defvar *sort-id-sort* 'void)
(defvar *string-sort* 'void)
(defvar *character-sort* 'void)
(defvar *axiom-sort* 'void)
(defvar *object-identifier-sort* 'void)
(defvar *object-sort* 'void)
(defvar *record-instance-sort* 'void)
(defvar *class-id-sort* 'void)
(defvar *record-id-sort* 'void)
(defvar *attribute-id-sort* 'void)
(defvar *attribute-sort* 'void)
(defvar *attribute-list-sort* 'void)
(defvar *attr-value-sort* 'void)
(defvar *message-sort* 'void)
(defvar *configuration-sort* 'void)
(defvar *acz-configuration-sort* 'void)
(defvar *sort_builtin* 'void)

(defvar sup-universal-sort-name nil)
(defvar sup-huniversal-sort-name nil)

;;;=============================================================================
;;; Builtin modules & operations.
;;; Operations in some builtin modules.
;;;=============================================================================

(defvar *system-standard-prelude* nil)
(defvar *system-soft-wired* nil)
(defvar *kernel-hard-wired-builtin-modules* nil)

;;; *SYSTEM-MODULE*
(defvar *system-module* nil)

;;; *CHAOS-MODULE*
;;; bounds built in module CHAOS.
(defvar *chaos-module* nil)

;;; *CHAOS-SORT-ORDER* holds the transitive closure of sort relations between
;;; builtin sorts. This is a value of slot 'sort-order' of builtin module
;;; CHAOS (buoud to global *chaos-module*).
;;;
(defvar *chaos-sort-order* 'void)

;;; *PARSER-SORT-ORDER*
(defvar *parser-sort-order* 'void)
  
;;; HARD-WIRED CHAOS MODULE
(defvar *system-object-module* nil)
(defvar *identifier-module* nil)
(defvar *universal-module* nil)
(defvar *parser-module* nil)
(defvar *qid-module* nil)

;;; Some operators & methods of CHAOS module.
;;; these are used for representing builtin constant and ill-formed terms
;;; at parsing time.
;;;
(defvar *builtin-method* nil)
(defvar *builtin-op* nil)
(defvar *partial-op* nil)
(defvar *void-op* nil)
(defvar *partial-method* nil)
(defvar *void-method* nil)
(defvar *type-err-op* nil)
(defvar *type-err-method* nil)
(defvar *op-err-op* nil)
(defvar *op-err-method* nil)

;;; TRUTH, BOOL & IDENTICAL Module
;;;-----------------------------------------------------------------------------
(defvar *TRUTH-module* 'void)
(defvar *BOOL-module* 'void)
(defvar *IDENTICAL-module* nil)

;;; basic operations in TRUTH & BOOL
;;;-----------------------------------------------------------------------------
(defvar *bool-true* 'void)
(defvar *bool-true-meth* 'void)
(defvar *bool-false* 'void)
(defvar *bool-false-meth* 'void)
(defvar *bool-and* 'void)
(defvar *bool-or* 'void)
(defvar *bool-not* 'void)
(defvar *sort-membership* 'void)
(defvar *bool-if* 'void)
(defvar *bool-equal* 'void)
(defvar *beh-equal* 'void)
(defvar *bool-nonequal* 'void)
(defvar *beh-eq-pred* 'void)

;;; RWL
;;;-----------------------------------------------------------------------------
(defvar *rwl-module* 'void)
(defvar *rwl-nat-star-sort* 'void)
(defvar *rwl-predicate* 'void)
(defvar *rwl-predicate2* 'void)

;;; basic operations in IDENTICAL.
;;;-----------------------------------------------------------------------------
(defvar *identical* 'void)
(defvar *nonidentical* 'void)

;;; constructors for record, object, attribute list.
;;;-----------------------------------------------------------------------------
(defvar *attribute-constructor* nil)
(defvar *attribute-list-constructor* nil)
(defvar *attribute-list-aux-variable* nil)
(defvar *object-reference-method* nil)
(defvar *object-constructor-method* nil)
(defvar *object-constructor-op* nil)
(defvar *record-constructor-method* nil)
(defvar *record-constructor-op* nil)
(defvar *void-object* nil)
(defvar *void-record* nil)

;;; ***************
;;; READER & PARSER_____________________________________________________________
;;; ***************

(declaim (special *parse-variables*
		  *fill-rc-attribute*
		  *lhs-attrid-vars*
		  *parsing-axiom-lhs*
		  *parse-lhs-attr-vars*)) ; binds variables during a parsing
					; process. 
(declaim (special  *reader-schema-env*	; current schema.
		   *reader-input*	; current token sequence.
		   ))

(defvar *fill-rc-attribute* nil)	; a flag, t if requires generalizing the
					; pattern of record/object terms.

(defvar *parsing-axiom-lhs* nil)
(defvar *parse-lhs-attr-vars* nil)
(defvar *lhs-attrid-vars* nil)

;;;
(defparameter .lisp-start-symbol. #\#)
(defvar *parse-variables* nil)
(defconstant parser-min-precedence  0)
(defconstant parser-max-precedence 127)
(defvar *reader-schema-env* nil)
(defvar *reader-current-schema* nil)
(defvar *reader-current-context* nil)
(defvar *reader-starting-position* nil)
(defvar *builtin-ast-dict* (make-hash-table :test #'equal))

;;; if t, parse speed can be greatly enhanced for many cases, but produces
;;; much less dignosis infos. this is a toplevel switch.
;;;
(defvar *fast-parse* t)

;;; *********
;;; MISC VARS___________________________________________________________________
;;; *********

;;; *INCLUDE-BOOL* 
(defvar *include-bool* t)
;;; *INCLUDE-RWTL"
(defvar *include-rwl* t)
;;;
(defvar *compile-lisp-rhs* t)
(defvar *running-with-tk* nil)
(defvar *sub-prompt* nil)
(defvar *consider-object* nil)

(defvar *auto-reconstruct* nil)

;;; *SAVE-DEFINITION*
(defvar *save-definition* t)

;;; *MODMORPH-NEW-MODULE*
(declaim (special *modmorph-new-module*))
(defvar *modmorph-new-module* nil)

;;; TIMEZOE
(defvar *time-zone* -9)

;;;
(defvar *allow-general-term-input* nil)
;;;
(defvar *chaos-libpath* nil)
;;;
(declaim (special *beh-proof-in-progress*))
(defvar *beh-proof-in-progress* nil)

;;; TRAM
(defvar *tram-path* "tram")
(defvar *tram-options* "")
(defvar *tram-builtin-modules*)
(defvar *tram-bool-modules*)

;;; EXEC
(declaim (type fixnum *cexec-limit*))
(defvar *cexec-trace* nil)
(defvar *cexec-limit* most-positive-fixnum)
(defvar *cexec-normalize* t)

;;; BUTILTIN
(defvar *compile-builtin-axiom* nil)
(defvar *bi-universal-operators* nil)

;;; OPEN/CLOSE WORLD -- not used.
;;; (declaim (special *closed-world*))
;; (defvar *closed-world* nil)

;;; DEBUG FLAGS
(defvar *rewrite-debug* nil)
(defvar *on-term-hash-debug* nil)
(defvar *on-axiom-debug* nil)
(defvar *beh-debug* nil)
(defvar *gen-rule-debug* nil)
(defvar *on-change-debug* nil)
(defvar *on-operator-debug* nil)
(defvar *on-sort-debug* nil)
(defvar *on-trs-debug* nil)
(defvar *on-import-debug* nil)
(defvar *on-modexp-debug* nil)
(defvar *on-view-debug* nil)
(defvar *match-debug* nil)
(defvar *module-dep-debug* nil)
(defvar *term-debug* nil)
(defvar *on-parse-debug* nil)
(defvar *regularize-debug* nil)
(defvar *on-tram-debug* nil)
(defvar *mel-debug* nil)
(defvar *check-import-mode* nil)

;;; EOF
