-- -*- Mode: Chaos -*-
-- $Id: std.mod,v 1.1.1.1 1998/05/16 11:27:50 sawada Exp $
-- lisp (in-package :chaos) ; we must in the package Chaos.
--> ===================================================================
**> Chaos standard prelude, including OBJ3 compatible builtin modules.
--> ===================================================================

lisp (setf *last-module* nil
           *current-module* nil)

-- built-ins must import BOOL explicitly if need.
lisp (setq *include-BOOL* nil)


" ******* The followings are now hard wired modules, defined in *******
  ******* chaos/boot/builtins.lisp
  *********************************************************************
--> ===================================================================
--> UNIVERSAL
--> In Chaos, Universal sort is hardwired and defined in 
--> the module CHAOS:UNIVERSAL. 
--> module UNIVERSAL{[ Universal ]}

--> ==================================================================
--> TRIV the trivial module
--> ------------------------------------------------------------------
sys:mod* TRIV { [ Elt ] }

--> ==================================================================
--> TRUTH-VALUE, TRUTH and BOOL
--> These are upward compatible with OBJ.
--> Extenstion: introduces `==>', .... *TODO*
--> ------------------------------------------------------------------

sys:mod! TRUTH-VALUE {
 -- import { pr(CHAOS:PARSER) }
 signature {
  [ Bool ] 
  op false : -> Bool { constr }
  op true : -> Bool { constr }
 }
}
sys:mod! TRUTH {
  pr (TRUTH-VALUE)
  signature {
    -- normal coditional: for strict and rigid type checking,
    -- this should be polymorphic and declared on demand where used.
    -- we need `polymorphic' attribute? or are there any more smart way?
    op if_then_else_fi : Bool Universal Universal -> Universal 
	{ strategy: (1 0) prec: 0 }
    -- _==_ & _=/=_ is for equational equal.		     
    op _==_ : Universal Universal -> Bool
       { strategy: (1 2 0) prec: 51 }
    op _=/=_ : Universal Universal -> Bool
       { strategy (1 2 0) prec: 51 }
    -- i want _==>_ for `rewriting logic', *TODO*
    -- op _==>_ : Universal Universal -> Bool
    --   { strategy (1 0) prec: 51 }
    -- op _=/=>_ : Universal Universal -> Bool
    --   { strategy (1 0) prec: 51 }
  }
  lisp (setup-truth)

  vars XU YU : Universal

  axioms {
    eq if true then XU else YU fi = XU .
    eq if false then XU else YU fi = YU .
   }

  axioms {
      eq XU == YU =
    	  #!!(coerce-to-Bool (term-equational-equal XU YU)) .
      eq XU =/= YU =
      	  #!!(coerce-to-Bool (not (term-equational-equal XU YU))) .
      -- eq XU ==> YU =
      --    #!!(coerce-to-Bool (term-equational-equal XU YU)) .
      -- eq XU =/=> YU =
      --   #!!(coerce-to-Bool (not (term-equational-equal XU YU))) .
  }

}

sys:mod! BOOL {
  pr (TRUTH)
  signature {
    op _and_ : Bool Bool -> Bool
       { assoc comm idr: true strat: (1 2 0) r-assoc prec: 55 }

    op _or_ : Bool Bool -> Bool
       { assoc comm idr: false strat: (1 2 0) r-assoc prec: 59 }

    op _xor_ : Bool Bool -> Bool
       { assoc comm idr: false strat: (1 2 0) r-assoc prec: 57 }

    op not_ : Bool -> Bool { prec: 53 }

    op _implies_ : Bool Bool -> Bool { r-assoc prec: 61 }
   }
  lisp (setup-Bool)
   
   axioms {
    vars A B : Bool 
    -- ---------------------------
    eq not true = false .
    eq not false = true .
    eq false and A = false .
    eq true or A = true .
    eq A implies B = (not A) or B .
    eq true xor true = false .
    }
}

--> ==================================================================
--> IDENTICAL : determines syntactical equality among terms.
--> Compatible with OBJ3 IDENTICAL.
--> ------------------------------------------------------------------
sys:mod! IDENTICAL {
  pr (BOOL)
  signature {
    -- syntactical equality.
    op _===_ : Universal Universal -> Bool { strategy: (0) prec: 51 }
    op _=/==_ : Universal Universal -> Bool { strategy: (0) prec: 51 }
  }
  axioms {
    vars XU YU : Universal
    -- ---------------------------------------------------
    eq XU === YU =
      #!! (coerce-to-Bool (term-is-similar? XU YU)) .
    eq XU =/== YU =
      #!! (coerce-to-Bool (not (term-is-similar? XU YU))) .
  }
}

lisp (progn (setup-IDENTICAL) 'done)

--> ==================================================================
--> NUBERS : NZNAT, NAT, INT, RAT, and FLOAT.
--> Similar to OBJ3 modules of the same name,
--> may lack some operations or contain additionals, to be checked.
--> Builtin constructors are defined in separate modules, e.g.,
--> NZNAT-VALUE for defining NzNat.
--> ------------------------------------------------------------------

sys:mod! NZNAT-VALUE {
  -- bsort : just the same as of OBJ3.
  bsort NzNat (is-NzNat-token create-NzNat prin1 is-nznat) 
}

sys:mod! NZNAT {
  protecting (NZNAT-VALUE)
  protecting (BOOL)

  signature {
    op _+_ : NzNat NzNat -> NzNat
    attr _+_ { assoc comm prec: 33 }
    op d : NzNat NzNat -> NzNat 
    attr d/2 { comm }
    op _*_ : NzNat NzNat -> NzNat
    attr _*_ { assoc comm idr: 1 prec: 31 }
    op _quo_ : NzNat NzNat -> NzNat
    attr _quo_ { l-assoc prec: 31 }
    op _<_ : NzNat NzNat -> Bool 
    attr _<_ { prec: 51 }
    op _<=_ : NzNat NzNat -> Bool 
    attr _<=_ { prec 51 }
    op _>_ : NzNat NzNat -> Bool
    attr _>_ { prec: 51 }
    op _>=_ : NzNat NzNat -> Bool
    attr _>=_{ prec: 51 }
    op s_ : NzNat -> NzNat 
    attr s_ { prec: 15 }
    }

  axioms {
    vars NN NM : NzNat
    -- ------------------------------------------------
    eq NN + NM = #! (+ NN NM) .
    eq d(NN,NM) = #! (if (= NN NM) 1 (abs (- NN NM))) .
    eq NN * NM = #! (* NN NM) .
    eq NN quo NM = #! (if (> NN NM) (truncate NN NM) 1) .
    eq NN < NM = #! (< NN NM) .
    eq NN <= NM = #! (<= NN NM) .
    eq NN > NM = #! (> NN NM) .
    eq NN >= NM = #! (>= NN NM) .
    eq s NN = #! (1+ NN) .
   }
}

sys:mod! NAT-VALUE {
  protecting(NZNAT-VALUE)
  bsort Nat (is-nat-token create-nat prin1 is-nat) 
  bsort Zero (is-zero-token create-zero prin1 is-zero) 
  [ NzNat < Nat, Zero < Nat ]
}

sys:mod! NAT {
  protecting (NZNAT)
  protecting (NAT-VALUE)

  signature {
    op _+_ : Nat Nat -> Nat { idr: 0 }
    op sd : Nat Nat -> Nat { comm }
    op _*_ : Nat Nat -> Nat 
    op _quo_ : Nat NzNat -> Nat
    op _rem_ : Nat NzNat -> Nat
    attr _rem_ { l-assoc prec: 31 }
    op _divides_ : NzNat Nat -> Bool
    attr _divides_ { prec: 51 }
    op _<_ : Nat Nat -> Bool
    op _<=_ : Nat Nat -> Bool
    op _>_ : Nat Nat -> Bool
    op _>=_ : Nat Nat -> Bool
    op s_ : Nat -> NzNat
    op p_ : NzNat -> Nat
    attr p_ { prec: 15 }
  }

  axioms {
    vars M N : Nat
    var NN : NzNat
    -- -----------------------------------------
    eq sd(M,N) = #! (abs (- M N)) .
    eq M + N = #! (+ M N) .
    eq N * 0 = 0 .
    eq M quo NN = #! (truncate M NN) .
    eq M rem NN = #! (rem M NN) .
    eq NN divides M = #! (= 0 (rem M NN)) .
    eq N < 0 = false .
    eq 0 < NN = true .
    eq NN <= 0 = false .
    eq 0 <= N = true .
    eq 0 > N = false .
    eq NN > 0 = true .
    eq 0 >= NN = false .
    eq N >= 0 = true .
    eq s 0 = 1 .
    eq p NN = #! (- NN 1) .
   }
}

sys:mod! INT-VALUE {
  protecting (NAT-VALUE)
  bsort Int (is-Int-token create-Int prin1 is-Int) 
  bsort NzInt (is-NzInt-token create-NzInt prin1 is-NzInt) 
  [ Nat < Int, NzNat < NzInt < Int ]
}

sys:mod! INT {
  protecting (NAT)
  protecting (INT-VALUE)

  signature {
    op _+_ : Int Int -> Int
    op _-_ : Int Int -> Int
    attr _-_ { r-assoc prec: 33 }
    op _*_ : Int Int -> Int
    --    op _*_ : NzInt NzInt -> NzInt
    op _quo_ : Int NzInt -> Int
    op _rem_ : Int NzInt -> Int
    op _divides_ : NzInt Int -> Bool
    op _<_ : Int Int -> Bool
    op _<=_ : Int Int -> Bool
    op _>_ : Int Int -> Bool
    op _>=_ : Int Int -> Bool
    op s_ : Int -> Int
    op -_ : Int -> Int { prec: 15 }
    op -_ : NzInt -> NzInt { prec: 15 }
  }

 axioms {
  vars I J : Int
  var NJ : NzInt
  -- ---------------------------------------
  eq - I = #! (- I) .
  eq I + J = #! (+ I J) .
  --  eq I - J = #! (- I J) .
  eq I - J = I + (- J) .
  eq I * J = #! (* I J) .
  eq I quo NJ = #! (truncate I NJ) .
  eq I rem NJ = #! (rem I NJ) .
  eq NJ divides I = #! (= 0 (rem I NJ)) .
  eq I < J = #! (< I J) .
  eq I <= J = #! (<= I J) .
  eq I > J = #! (> I J) .
  eq I >= J = #! (>= I J) .
  eq s I = 1 + I .
 }
}

sys:mod! RAT-VALUE {
  protecting (INT-VALUE)
  bsort Rat (is-Rat-token create-Rat RAT-print rationalp) 
  bsort NzRat (is-NzRat-token create-NzRat RAT-print is-NzRat) 
  [ Int < Rat, NzInt < NzRat < Rat ]
 }

sys:mod! RAT {
  protecting (INT)
  protecting (RAT-VALUE)
  signature {
    op -_  : Rat -> Rat { prec:  15 }
    op -_  : NzRat -> NzRat { prec: 15 }
    op _+_ : Rat Rat -> Rat
    op _-_ : Rat Rat -> Rat
    op _*_ : Rat Rat -> Rat
    -- op _*_ : NzRat NzRat -> NzRat
    op _/_ : Rat NzRat -> Rat
    attr _/_ { l-assoc prec: 31 }
    op _/_ : NzRat NzRat -> NzRat
    op _rem_ : Rat NzRat -> Rat
    op _<_ : Rat Rat -> Bool
    op _<=_ : Rat Rat -> Bool
    op _>_ : Rat Rat -> Bool
    op _>=_ : Rat Rat -> Bool
  }

  axioms {
    vars R S : Rat
    var NS : NzRat
    -- ---------------------------------
    eq R + S = #! (+ R S) .
    eq - R = #! (- R) .
    -- eq R - S = #! (- R S) .
    eq R - S = R + (- S) .
    eq R * S = #! (* R S) .
    eq R / NS = #! (/ R NS) .
    eq R rem NS = #! (rem R NS) .
    eq R < S = #! (< R S) .
    eq R <= S = #! (<= R S) .
    eq R > S = #! (> R S) .
    eq R >= S = #! (>= R S) .
  }
}

sys:mod! FLOAT-VALUE {
    bsort Float (is-Float-token create-Float print-Float is-Float) 
}

sys:mod! FLOAT {
 pr (FLOAT-VALUE)
 pr (BOOL)
 signature {
    op _+_ : Float Float -> Float
    attr _+_ { assoc comm prec: 33 }
    op _-_ : Float Float -> Float
    attr _-_ { r-assoc prec: 33 }
    op _*_ : Float Float -> Float 
    attr _*_ { assoc comm prec: 31 }
    op _/_ : Float Float -> Float
    attr _/_ { l-assoc prec: 31 }
    op _rem_ : Float Float -> Float
    attr _rem_ { l-assoc prec: 31 }
    op exp : Float -> Float
    op log : Float -> Float
    op sqrt : Float -> Float
    op abs : Float -> Float
    op sin : Float -> Float
    op cos : Float -> Float
    op atan : Float -> Float
    op pi : -> Float 
    op _<_ : Float Float -> Bool
    attr _<_ { prec: 51 }
    op _<=_ : Float Float -> Bool
    attr _<=_ { prec: 51 }
    op _>_ : Float Float -> Bool
    attr _>_ { prec: 51 }
    op _>=_ : Float Float -> Bool 
    attr _>=_ { prec: 51 }
    op _=[_]_ : Float Float Float -> Bool
    attr _=[_]_ { prec: 51 }
    }

 axioms {
    vars X Y Z : Float
    -- -----------------------------------------
    eq X + Y = #! (+ X Y) .
    eq X - Y = #! (- X Y) .
    eq X * Y = #! (* X Y) .
    eq X / Y = #! (/ X Y) .
    eq X rem Y = #! (rem X Y) .
    eq exp(X) = #! (exp X) .
    eq log(X) = #! (log X) .
    eq sqrt(X) = #! (sqrt X) .
    eq abs(X) = #! (abs X) .
    eq sin(X) = #! (sin X) .
    eq cos(X) = #! (cos X) .
    eq atan(X) = #! (atan X) .
    eq pi = #! pi .
    eq X < Y = #! (< X Y) .
    eq X <= Y = #! (<= X Y) .
    eq X > Y = #! (> X Y) .
    eq X >= Y = #! (>= X Y) .
    eq (X =[ Z ] Y) = #! (< (abs (- X Y)) Z) .
    }
}

--> ==================================================================
--> QID : Quoted IDentifier, symbols starting with ' character.
--> for OBJ3 compatibility.
--> ------------------------------------------------------------------
sys:mod! QID {
 -- pr (CHAOS:PARSER)
 bsort Id (is-qid-token create-qid print-qid is-qid) 
}

--> ==================================================================
--> QIDL : extended QID with _<_ operation.
--> also for OBJ3 compatibility.
--> ------------------------------------------------------------------
sys:mod! QIDL {
 pr (QID)
 pr (BOOL)
 signature {
   op _<_ : Id Id -> Bool { prec: 51 }
 }
 axioms {
  vars X Y : Id
  eq X < Y = #! (string< X Y) .
 }
}

--> ==================================================================
--> CHARACTER
--> basic operations on hardwired builtin sort Character.
--> ------------------------------------------------------------------
sys:mod! CHARACTER {
 pr(BOOL)
 pr(NAT)
 signature {
    bsort Character (Is-character-token create-character
	             print-character is-character)
    op char-code : Character -> Nat
    op code-char : Nat -> Character
    op upcase : Character -> Character
    op downcase : Character -> Character
    op graphic-char-p : Character -> Bool
    op alpha-char-p : Character -> Bool
    op upper-case-p : Character -> Bool
    op lower-case-p : Character -> Bool
    op both-case-p  : Character -> Bool
    op digit-char-p : Character -> Bool
    op alphanumericp : Character -> Bool
    op char< : Character Character -> Bool
    op char> : Character Character -> Bool
    op char<= : Character Character -> Bool
    op char>= : Character Character -> Bool
    op char= : Character Character -> Bool
 }
 axioms {
    vars C C2 : Character
    var N : Nat
    -- -----------------------------------------
    eq char-code(C) = #! (char-code C) .
    cq code-char(N) = #! (code-char N) :if N <= 255 .
    eq upcase(C) = #! (char-upcase C) .
    eq downcase(C) = #! (char-downcase C) .
    eq graphic-char-p(C) = #! (graphic-char-p C) .
    eq alpha-char-p(C) = #! (alpha-char-p C) .
    eq upper-case-p(C) = #! (upper-case-p C) .
    eq lower-case-p(C) = #! (lower-case-p C) .
    eq both-case-p(C) = #! (both-case-p C) .
    eq digit-char-p(C) = #! (digit-char-p C) .
    eq alphanumericp (C) = #! (alphanumericp C) .
    eq char=(C,C2) = #! (char= C C2) .
    eq char<(C, C2)  = #! (char< C C2) .
    eq char>(C, C2) = #! (char> C C2) .
    eq char<=(C, C2) = #! (char<= C C2) .
    eq char>=(C, C2) = #! (char>= C C2) .
    }
}



lisp (progn (install-character) 'done)

--> ==================================================================
--> STRING
--> basic operations on hardwired builtin sort String.
--> ------------------------------------------------------------------
sys:mod! STRING {
 pr(CHARACTER)
 signature {
    [ String ]
    op schar : String Nat -> Character
    op _++_ : String String -> String { r-assoc constr }
    op string : Character -> String
    op upcase : String -> String
    op downcase : String -> String
    op capitalize : String -> String
    op substring : String Nat Nat -> String    -- substring(String, from, to)
    op substring : String Nat -> String        -- substring(String, from)
    op rsubstring : String Nat -> String       -- substring(String, to)
    op length : String -> Nat
    op string= : String String -> Bool
    op string< : String String -> Bool
    op string> : String String -> Bool
    op string<= : String String -> Bool
    op string>= : String String -> Bool
    }
 axioms {
    vars S S1 S2 : String
    var C : Character
    vars N FROM TO : Nat
    -- -----------------------------------------------------------------
    eq S ++ S2 = #! (concatenate 'string S S2) .
    eq string(C) = #! (string C) .
    eq schar(S, N) = #! (schar S N) .
    eq upcase(S) = #! (string-upcase S) .
    eq downcase(S) = #! (string-downcase S) .
    eq capitalize(S) = #! (string-capitalize S) .
    eq substring(S, FROM, TO) = #! (subseq S FROM TO) .
    eq substring(S, N) = #! (subseq S N).
    eq rsubstring(S, N) = #! (subseq S 0 N) .
    eq length(S) = #! (length S) .
    eq string=(S1, S2) = #! (string= S1 S2) .
    eq string<(S1, S2) = #! (string< S1 S2) .
    eq string>(S1, S2) = #! (string> S1 S2) .
    eq string<=(S1, S2) = #! (string<= S1 S2) .
    eq string>=(S1,S2) = #! (string>= S1 S2) .
    }
}

lisp (progn (install-string) 'done)

--> ==================================================================
--> PROPC: basic propositional calculus. 
--> This is stolen from OBJ3 example programs, and adapted to Chaos.
--> ------------------------------------------------------------------
sys:mod! PROPC {
  pr (CHAOS:IDENTIFIER)
  pr (TRUTH)
  -- Prop is for proposition.
  -- Atomic proposition is either a truth value or an Identifier.
  [ Prop, Identifier Bool < Prop ]

  -- canonical form : and & xor
  -- all other logical operations are represented in
  -- terms of `and' and `xor'.
  signature {
    op _and_ : Prop Prop -> Prop { assoc comm prec: 55 constr }
    op _xor_ : Prop Prop -> Prop { assoc comm prec: 57 constr }
    }

  -- other operations:
  signature {
    op _or_ : Prop Prop -> Prop { assoc comm prec: 59 }
    op not_ : Prop -> Prop { prec: 53 }
    op _->_ : Prop Prop -> Prop { prec: 61 }
    op _<->_ : Prop Prop -> Prop { assoc prec: 63 }
    }

  axioms { 
    vars p q r : Prop
    -- ----------------------------------------------
    eq p and false = false .
    eq p and true = p .
    eq p and p = p .
    eq p xor false = p .
    eq p xor p = false .
    eq p and (q xor r) = (p and q) xor (p and r) .
    -- 
    eq p or q = (p and q) xor p xor q .
    eq not p = p xor true .
    eq p -> q = (p and q) xor p xor true .
    eq p <-> q = p xor q xor true .
    }
}

--> ==================================================================
--> RECORD Instance & OBJECT :
--> These provide basic building blocks for record instance, objects,
--> and state configuration for representing distributed status of
--> object and messages configuration.
--> ------------------------------------------------------------------

** Attribute Value pairs.
-- used for constructing Object & RecordInstance
-- -------------------------------------------------------------------

sys:mod! AVPAIR {
 -- AttrId : attribute id (slot name).
 -- AttrValue : attribute's value.
 -- Attribute : pair of attribute id & its value.
 -- Attributes : sequence of attribute-value paris.
 [ AttrId, AttrValue, Attribute < Attributes ]

 -- Attribute
 op (_=_) : AttrId AttrValue -> Attribute { prec: 49 constr }
 op attr-null : -> Attribute

 -- Attributes
 op _,_ : Attributes Attributes -> Attributes
   { associative commutative id: attr-null prec: 63 constr }

}

** RECORD CONSTRUCTION KIT.
-- -------------------------------------------------------------------
sys:mod! RECORD-STRUCTURE {
 pr (AVPAIR)
 [ RecordId RecordInstance ]
 -- Record Constructor.
 -- a record instance consists of its record sort name (RecordId),
 -- and its attribute-value pairs (Attributes).
 op *VoidRecord* : -> RecordInstance
 op (_{_}) : RecordId Attributes -> RecordInstance { constr }
 op (_{}) : RecordId -> RecordInstance
 attr (_{_}) { prec: 0 }
 attr(_{}) { prec: 0 }
}

sys:mod! OBJECT-ID {
  pr (CHAOS:IDENTIFIER)
  -- We predefine Identifier as object identifier(ObjectId).
  [ Identifier < ObjectId ]
}

lisp (progn (install-object-id) 'done)
	  
** OBJECT CONSTRUCTION KIT.
-- -------------------------------------------------------------------
sys:mod! OBJECT {
  import {
    pr (OBJECT-ID)                       -- object identifier
    pr (BOOL)
    pr (AVPAIR)				 -- a set of attribute-value pairs
  }

  signature {
     -- ClassId : name of class sort.
     -- Object  : instance of class sort.
     -- Message : message to objects.
     [ ClassId, Message, Object ]
     -- Object Constructors.
     --  -  an object consists of its unique indentifier (ObjectId),
     --  -  its class (ClassId), and its attribute-value pairs (Attributes).

     op (<_:_|_>) : ObjectId ClassId Attributes -> Object { constr }
     attr (<_:_|_>) { prec: 0 }
     op (<_:_>) : ObjectId ClassId -> Object
     attr (<_:_>) { prec: 0 }
     op (<_>) : ObjectId -> Object
     attr (<_>) { prec: 0 }
     op *VoidObject* : -> Object
     op delete_ : Object -> Bool
  }
  
  axioms {
	 -- some primitive axioms.
    eq [ find-instance ] : 
       < Id:ObjectId : C:ClassId > 
       = #!! (or (find-instance Id (term-sort *self*)) *void-object*)  .
     eq [ find-instnace-2 ] :
       < Id:ObjectId >
      = #!! (or (find-instance Id) *void-object*) .
     rule [ delete-instance ] :
       delete O:Object => #!! (if (delete-instance O) *bool-true* *bool-false*) .
  }
}


** CONFIGURATION
-- -------------------------------------------------------------------
sys:mod! STATE-CONFIGURATION {
 pr (OBJECT)
 -- Configuration represents the world in which objects and messages
 -- plays. 
 [ Configuration ]
}

** ACZ-CONFIGURATION
-- represents a kind of distributes state of Messages and OBjects.
-- this is a standard state configuration.
-- -------------------------------------------------------------------
sys:mod! ACZ-CONFIGURATION {
  extending (STATE-CONFIGURATION) 
  [ Object Message < ACZ-Configuration < Configuration ]
  -- ACZ-Configuration:
  -- a list of objects and messages constructed by an associative 
  -- and commutative constucting operator (_ _) .
  -- the universe consists of list of objects and messages.
  op acz-empty : -> ACZ-Configuration { constr }
  op __ : ACZ-Configuration ACZ-Configuration -> ACZ-Configuration 
	  { assoc comm id: acz-empty constr }
}


**********************************************************************
"

**
** Now only followings are included in system standard prelude.
** 
--> ===================================================================
--> OBJ3 builtin modules:  ERR, BUILT-IN, LISP
--> don't know how these modules are used in OBJ.
--> -------------------------------------------------------------------
sys:mod! ERR {
 bsort Err
   (is_Err_token
    create_Err
    print_Err
    is_Err)
}

lisp (progn (install-err) 'done)
protect ERR

sys:mod! BUILT-IN {
 bsort Built-in (is_Builtin_token create_Builtin print_Builtin is_Builtin)
}

lisp (progn (install_builtin) 'done)
protect BUILT-IN

sys:mod! LISP {
 bsort Lisp (is_Lisp_token create_Lisp print_Lisp is_Lisp)
}
protect BUILT-IN

--> ==================================================================
--> LAST-TERM
--> For OBJ3 compatibility. Currently not used in Chaos.
--> ------------------------------------------------------------------
sys:mod! LAST-TERM {
  -- protecting (CHAOS:UNIVERSAL)
  protecting (TRUTH-VALUE)
  signature {
    op last-term-undefined : -> Universal
    op last-term : -> Universal
    }
  axioms {
    eq last-term = last-term-undefined .
  }
}

lisp (progn (install-last-term) 'done)
protect LAST-TERM

--> ==================================================================
--> 2TUPLE, 3TUPLE, 4TUPLE
--> for OBJ3 compatibility.
--> ------------------------------------------------------------------
sys:mod! 2TUPLE [ C1 :: TRIV, C2 :: TRIV ] {
  signature {
   [ 2Tuple ]
   op <<_;_>> : Elt.C1 Elt.C2 -> 2Tuple {constr}
   op 1*_ : 2Tuple -> Elt.C1 
   op 2*_ : 2Tuple -> Elt.C2
  }
  axioms {
   var e1 : Elt.C1 
   var e2 : Elt.C2
   -- -----------------------------
   eq 1* << e1 ; e2 >> = e1 .
   eq 2* << e1 ; e2 >> = e2 .
  }
}
protect 2TUPLE

sys:mod! 3TUPLE[C1 :: TRIV, C2 :: TRIV, C3 :: TRIV] {
  signature {
   [ 3Tuple ]
   op <<_;_;_>> : Elt.C1 Elt.C2 Elt.C3 -> 3Tuple {constr}
   op 1*_ : 3Tuple -> Elt.C1 
   op 2*_ : 3Tuple -> Elt.C2 
   op 3*_ : 3Tuple -> Elt.C3 
  }
  axioms {
   var e1 : Elt.C1 
   var e2 : Elt.C2
   var e3 : Elt.C3
   -- -----------------------------
   eq 1* << e1 ; e2 ; e3 >> = e1 .
   eq 2* << e1 ; e2 ; e3 >> = e2 .
   eq 3* << e1 ; e2 ; e3 >> = e3 .
  }
}
protect 3TUPLE

sys:mod! 4TUPLE[C1 :: TRIV, C2 :: TRIV, C3 :: TRIV, C4 :: TRIV] {
  signature {
   [ 4Tuple ]
   op <<_;_;_;_>> : Elt.C1 Elt.C2 Elt.C3 Elt.C4 -> 4Tuple {constr}
   op 1*_ : 4Tuple -> Elt.C1 
   op 2*_ : 4Tuple -> Elt.C2 
   op 3*_ : 4Tuple -> Elt.C3 
   op 4*_ : 4Tuple -> Elt.C4 
  }
  axioms {
   var e1 : Elt.C1 
   var e2 : Elt.C2 
   var e3 : Elt.C3 
   var e4 : Elt.C4 
   -- ----------------------------------
   eq 1* << e1 ; e2 ; e3 ; e4 >> = e1 .
   eq 2* << e1 ; e2 ; e3 ; e4 >> = e2 .
   eq 3* << e1 ; e2 ; e3 ; e4 >> = e3 .
   eq 4* << e1 ; e2 ; e3 ; e4 >> = e4 .
  }
}
protect 4TUPLE
  
--
-- important initializations:
--
lisp (setq *include-BOOL* t)
lisp (setf *last-module* nil
         *current-module* nil)
lisp (progn (install-prelude) 'done)

eof
