;;;-*- Mode:LISP; Package: COMMON-LISP-USER; Base:10; Syntax:Common-Lisp -*-
;;; $Id: make-cafeobj.lisp,v 1.6 1998/06/27 04:01:17 sawada Exp $

;;;*****************************************************************************
;;; Template for making CafeOBJ interpreter.
;;;*****************************************************************************

(eval-when (eval load)
  (unless (find-package :common-lisp)
    (rename-package :lisp :common-lisp
		    (union '("CL" "LISP")
			   (package-nicknames (find-package :lisp)) :test #'string=)))
  (unless (find-package :common-lisp-user)
    (rename-package :user :common-lisp-user
		    (union '("CL-USER" "USER")
			   (package-nicknames (find-package :user)) :test #'string=)))
  )

#+LUCID (in-package "user")
#+Excl (in-package :user)
#+(or :ccl CMU) (in-package :common-lisp-user)
#+GCL (in-package :user)
#+GCL (setf compiler::*compile-ordinaries* t)
(unless (find-package :chaos)
  (make-package :chaos))

;;; BEGIN SITE SPECIFIC ------------------------------------------------------
(defvar *chaos-root*)
(eval-when (eval load)
  (setq *chaos-root* XCHAOS_ROOT_DIR))

(defvar chaos::*cafeobj-install-dir* nil)

(eval-when (eval load)
  (setq chaos::*cafeobj-install-dir* XINSTALL_DIR)
  )
;;; END SITE SPECIFIC --------------------------------------------------------

#-:mk-defsystem
(eval-when (eval load)
  #-CCL (load (concatenate 'string *chaos-root* "/defsystem"))
  ;; patch by t-seino@jaist.ac.jp
  #+CCL (load (concatenate 'string *chaos-root* ":defsystem"))
)

#+GCL
(defun make-exec-image (path)
  (setq chaos::-cafeobj-load-time- (chaos::get-time-string))
  (chaos::set-cafeobj-standard-library-path)
  (setq *chaos-vergine* t)
  (defun system::top-level () (funcall 'chaos::cafeobj-top-level))
  (si::set-up-top-level)
  (system::save-system path)
  (bye))

#+CMU
(defun make-exec-image (path)
  (setq chaos::-cafeobj-load-time- (chaos::get-time-string))
  (chaos::set-cafeobj-standard-library-path)
  (setq *chaos-vergine* t)
  (ext:save-lisp path
		 :purify t
		 :init-function 'chaos::cafeobj-top-level
		 :print-herald nil
		 )
  )

#+LUCID
(defun make-exec-image (path)
  (setq chaos::-cafeobj-load-time- (chaos::get-time-string))
  (chaos::set-cafeobj-standard-library-path)
  (setq *chaos-vergine* t)
  (disksave path
	    :full-gc t
	    :restart-function 'chaos::cafeobj-top-level))

#+:ccl
(defun make-exec-image (path)
  (setq chaos::-cafeobj-load-time- (chaos::get-time-string))
  (chaos::set-cafeobj-standard-library-path)
  (setq *chaos-vergine* t)
  ;; patch by t-seino@jaist.ac.jp
  (save-application path :toplevel-function 'chaos::cafeobj-top-level)
  :size '(6144000 4196000))

#+:ALLEGRO-V5.0
(require :build)

#+:ALLEGRO
(defun make-exec-image (path)
  (setq chaos::-cafeobj-load-time- (chaos::get-time-string))
  (chaos::set-cafeobj-standard-library-path)
  (setq *chaos-vergine* t)
  (setq excl:*restart-app-function* 'chaos::cafeobj-top-level)
  #||
  (setq excl:*restart-init-function*
    #'(lambda () (setq excl:*print-startup-message* nil)))
  ||#
  (setq excl:*print-startup-message* nil)
  (setq excl::.dump-lisp-suppress-allegro-cl-banner. t)
  (dumplisp :name path :suppress-allegro-cl-banner t)
  )

(defvar chaos::*compile-builtin-axiom* nil)

(defun make-cafeobj (&optional chaos-root)
  (when chaos-root
    (setf *chaos-root* chaos-root))
  (setq chaos::*compile-builtin-axiom* t)
  (mk::operate-on-system :chaosx :compile)
  (in-package :chaos)
  (setq chaos::*compile-builtin-axiom* nil)
  ;;
  (make-exec-image
   (concatenate 'string *chaos-root*
		#+GCL "/xbin/cafeobj.exe"
		#+Allegro "/xbin/cafeobj.acl"
		#+CMU "/xbin/cafeobj.core"
		;; patch by t-seino@jaist.ac.jp
		#+CCL ":xbin:cafeobj.exe"))
  )

(eval-when (eval load)
  (make-cafeobj *chaos-root*)
  )
;;; EOF
