;;;-*- Mode:LISP; Package:CHAOS; Base:10; Syntax:Common-lisp -*-
;;; $Id: basics.lisp,v 1.2 1999/02/14 03:50:13 sawada Exp $
(in-package :chaos)
#|==============================================================================
				 System: CHAOS
				Module: thstuff
			       File: basics.lisp
==============================================================================|#
#-:chaos-debug
(declaim (optimize (speed 3) (safety 0) #-GCL (debug 0)))
#+:chaos-debug
(declaim (optimize (speed 1) (safety 3) #-GCL (debug 3)))

;;; === Common Basic Functions for APPLY family
;;;     ------------------------------------------------------------------------

;;; *****
;;; UTILS
;;; *****

(defun check-apply-context (mod)
  (unless (check-$$term-context mod)
    (with-output-chaos-warning ()
      (format t "the target term : ")
      (print-chaos-object $$term)
      (print-next)
      (format t "isn't proper in the context : ")
      (print-chaos-object mod)
      (print-next)
      (format t "please re-`start' with supplying new one!")
      (throw 'apply-context-error nil))))

(defun command-display ()
  (if $$action-stack
      (format t "~&-- condition(~s) " (length $$action-stack))
      (format t "~&result "))
  (disp-term $$term))

(defun command-final ()
  (when $$action-stack
    ;; $$action-stack == list of
    ;;      ($$term term rule condition rhs-instance sort)
    ;;         0      1   2      3          4
    (if (term-is-similar? $$term *bool-true*)
	(progn
	  (command-display)
	  (format t "~&-- condition is satisfied, applying rule")
	  (format t "~&-- shifiting focus back to previous context")
	  (let ((cur (car $$action-stack)))
	    (setq $$term (car cur))
	    (term-replace (nth 1 cur) (nth 4 cur))
	    (setq $$action-stack (cdr $$action-stack)))
	  t)
	(if (term-is-similar? $$term *bool-false*)
	    (progn
	      (command-display)
	      (format t "~&-- condition is not satisfied, rule not applied")
	      (format t "~&-- shifting focus back to previous context")
	      (setq $$term (caar $$action-stack))
	      (setq $$action-stack (cdr $$action-stack))
	      t)
	    nil))))

(defun disp-term (x)
  (with-in-module (*last-module*)
    (term-print x)
    (princ " : ")
    (print-sort-name (term-sort x) *current-module*)))

(defun disp-term* (x)
  (term-print x)
  (princ " : ")
  (print-sort-name (term-sort x) *current-module*))

;;;
;;; apply-help
;;;
(defun apply-help ()
  (format t "~&Apply a selected rule, possibly with an instantiation,")
  (format t " to selected subterm(s).")
  (format t "~&Syntax:")
  (format t "~&  apply { reduction | red | print | bred | exec | <RuleSpec> [ <VarSubst> ] }")
  (format t "~&        { at | within } <Selectors>")
  (format t "~&  <RuleSpec>  ::= [ + | - ] [<ModId>.]<RuleId>")
  (format t "~&  <RuleId>    ::= <Natural> | <Identifier>")
  (format t "~&  <VarSubst>  ::= with <Variable> = <Term> { , <Variable> = <Term> }*")
  (format t "~&  <Selectors> ::= top | term | subterm | <Selector> { of <Selector> }* .")
  (format t "~&  <Selector>  ::= (<Natural>+) | `[' <Natural> [ .. <Natural> ] `]' |")
  (format t "~&                  `{' <Natural> {, <Natural>}* `}'")
  )

;;; *******************
;;; COMPUTING SELECTION
;;; *******************

(defun str-to-int (x)
  (if (equal "" x)
      0
      (read-from-string x)))

(defun !make-right-associative (method subs)
  (if (null (cdr subs)) (car subs)
    (make-right-assoc-normal-form method subs)
  ))

;;; <OccurSelection>
;;; ________________
(defun compute-occur-selection (sort term occs)
  (let ((cursrt sort) (cur term))
    (dolist (i occs)
      (if (and (<= 1 i) (<= i (length (term-subterms cur))))
	  (progn
	    (setq cursrt (nth (1- i) (method-arity (term-head cur))))
	    (setq cur (nth (1- i) (term-subterms cur))))
	  (with-output-chaos-warning ()
	    (format t "no such occurrenct, occurrence ~a is not correct for term :"
		    occs)
	    (print-next)
	    (term-print term)
	    (print-next)
	    (format t "ignoring it.")
	    (return-from compute-occur-selection (values sort term)))))
    (values cursrt cur) ))

;;; <SubseqSelection>
;;; _________________
;;; compute-subseq-selection : sort term m n
;;;
;;; <SubseqSelection> ::= [ m .. n ] | [ m ]
;;;  - case [ m ]  --> [ m .. m ]
;;;
(defun compute-subseq-selection (sort term m1 n1)
  (let ((m (1- m1))			; *note* sequence specifier is 1 origin.
	(n (1- n1)))
    (if (term-is-variable? term)
	(with-output-chaos-warning ()
	  (format t "found variable in selection of subsequence selection : ")
	  (term-print term)
	  (values sort term))
	(let ((method (term-head term)))
	  ;; [ ... ] is meaningful only for associative operators.
	  (if (method-is-associative method)
	      (let ((lst (list-assoc-subterms term method)))
		(if (and (<= m n) (<= 1 m1) (<= n1 (length lst)))
		    (if (or (< 1 m1) (< n1 (length lst)))
			(let ((res (!make-right-associative
				    method
				    (subseq lst m (1+ n)))))
			  (term-replace term
					(!make-right-associative
					 method
					 (append (subseq lst 0 m)
						 (list res)
						 (subseq lst (1+ n)))))
			  (values (car (method-arity method)) res))
			(values sort term))
		    (with-output-chaos-warning ()
		      (format t "selection [~a .. ~a] is out of range for term :~%    "
			      m1 n1)
		      (term-print term)
		      (print-next)
		      (format t "selected the whole subterms instead.")
		      (values sort term)
		      )))
	      (with-output-chaos-warning ()
		(format t "found non-associative operator in selection of subsequence slection : ~%    ")
		(print-chaos-object method)
		(values sort term)))))))

;;; <SubsetSelection>
;;; _________________

(defun compute-subset-selection (sort term occs)
 (if (term-is-variable? term)
     (with-output-chaos-warning ()
       (format t "found variable in subset selection ~a: " occs)
       (term-print term)
       (print-next)
       (format t "ignoring the selection and select whole subterms instead.")
       (values sort term))
     (let ((method (term-head term)))
       (if (and (method-is-associative method)
		(method-is-commutative method))
	   (let ((lst (list-AC-subterms term method)))
	     (let ((len (length lst))
		   (sel nil)
		   (rest nil)
		   (err nil))
	       (dolist (i occs)
		 (let ((n (1- i)))
		   (if (and (<= 0 n) (< n len))
		       (let ((tl (nthcdr n lst)))
			 (when (car tl) (push (car tl) sel))
			 (rplaca tl nil))
		       (push i err))))
	       (dolist (x lst) (when x (push x rest)))
	       (when err
		 (with-output-chaos-warning ()
		   (princ "found out of range in selection of subterms")
		   (print-next)
		   (format t "ignoring these selections : ~a" err)))
	       (if (null rest)
		   (values sort term)
		   (let ((res (!make-right-associative method (nreverse sel))))
		     (term-replace
		      term
		      (!make-right-associative method (cons res (reverse rest))))
		     (values (car (method-arity method)) res)
		     ))))
	   (with-output-chaos-warning ()
	     (princ "subset selection is meaningful only for associative and commuteative operators,")
	     (print-next)
	     (format t "but : ")
	     (print-chaos-object method)
	     (princ " is not.")
	     (print-next)
	     (princ "ignoreing the selection and select whole subterms instead.")
	     (values sort term)))
       )))

;;; compute-set-ocs
;;;  ("{" "1" "," "2" "," "4" "}") --> (1 2 4)
;;;
(defun compute-set-ocs (x)
  (let ((res nil)
	(val nil))
    (setq x (cdr x))
    (do ((elt x (cddr elt)))
	((endp elt) (nreverse res))
      (setq val (str-to-int (car elt)))
      (pushnew val res))))

;;; top-level interface
;;; *******************
(defvar *selection-target* nil)
(declaim (special *selection-target*))

(defun compute-selection (tm sel)
  (let ((*selection-target* tm))
    (!compute-selection *cosmos*
			tm
			(if (consp sel)
			    (if (equal (car (last sel)) ".")
				(butlast sel)
				sel)
			    sel))))

;;; !compute-selection
;;; the main computing selections
;;;
(defun !compute-selection (sort tm sel)
  ;; no selection
  (unless sel (return-from !compute-selection
		(if (not (term-eq $$term *selection-target*))
		    (values sort *selection-target*)
		    (values sort $$term))))
  ;; whole term
  (when (memq sel '(:term :top))
    (return-from !compute-selection (values sort $$term)))
  ;; subterm
  (when (eq sel :subterm)
    (if $$subterm
	(return-from !compute-selection (values sort $$subterm))
	(progn
	  (with-output-chaos-warning ()
	    (format t "no subterm is specified yet, please `choose' some.")
	    (print-next)
	    (format t "selected the whole term."))
	  (values sort $$term))))
  ;; 
  (!compute-selection-aux sort tm (cons (car sel) (cadr sel))))

(defun !compute-selection-aux (sort tm sel)
  (unless sel (return-from !compute-selection-aux
		(if (not (term-eq *selection-target* $$term))
		    (values sort *selection-target*)
		    (values sort $$term))))
  ;; <Selection> of <Selection>
  (when (equal "of" (car sel))
    (return-from !compute-selection-aux
      (!compute-selection-aux sort tm (cdr sel))))
  ;; 
  (case-equal (caar sel)
     ("("
      ;; occur selection
      (when (equal ")" (cadr (car sel)))
	(return-from !compute-selection-aux
	  (!compute-selection-aux sort tm (cdr sel))))
      (multiple-value-bind (s1 t1)
	  (!compute-selection-aux sort tm (cdr sel))
	(return-from !compute-selection-aux
	  (compute-occur-selection s1 t1
				   (mapcar #'str-to-int (cadr (car sel)))))))
     ("["
       ;; subseq selection { [ m .. n ] | [ m ] }
       (let ((i1 (str-to-int (cadr (car sel)))))
	 (multiple-value-bind (s1 t1)
	     (!compute-selection-aux sort tm (cdr sel))
	   (compute-subseq-selection s1
				     t1
				     i1
				     (if (equal "]" (nth 2 (car sel)))
					 ;; case [ m ]
					 i1 
					 ;; case [ m .. n ]
					 (str-to-int (cadr (nth 2 (car sel)))))))))
     ("{"
      ;; subset selection
      (multiple-value-bind (s1 t1)
	  (!compute-selection-aux sort tm (cdr sel))
	(compute-subset-selection s1 t1 (compute-set-ocs (car sel)))))
     (t (break "SNARK: !compute-selection"))))

;;; **************
;;; Utils on TERMS 
;;; **************

(defun @copy-term (term)
  (simple-copy-term term))

(defun @matcher (pat term)
  (if (term-is-variable? pat)
      (if (sort<= (term-sort term) (variable-sort pat)
		  (module-sort-order *current-module*))
	  (values nil (list (cons pat term)) nil nil)
	  (values nil nil t nil))
      (if (term-is-lisp-form? pat)
	  (values nil nil t nil)
	  (first-match pat term))))

(defun @test-rule-direct (rul term)
  (multiple-value-bind (gs sub no eeq)
      (@matcher (axiom-lhs rul) term)
    (declare (ignore gs sub eeq))
    (null no)))

(defvar *-inside-apply-with-extensions-* nil)

(defun @test-rule (rule term)
  (multiple-value-bind (gs sub no-match eeq)
      (@matcher (axiom-lhs rule) term)
    (declare (ignore gs sub eeq))
    (if (and no-match
	     (and *-inside-apply-with-extensions-*
		  (not (or (term-is-variable? term)
			   (term-is-builtin-constant? term)))
		  (method-is-of-same-operator (term-head (axiom-lhs rule))
					      (term-head term))))
	(@test-rule-extensions rule term)
	(null no-match)
	)))

(defun @make-ac-pattern (top term)
  (let ((newvar (make-variable-term *cosmos* 'ac-pat)))
    (make-right-assoc-normal-form top
				  (cons newvar
					(list-assoc-subterms
					 term
					 (term-head term))))))
(defun @make-a-patterns (top term)
  (let ((new-var1 (make-variable-term *cosmos* 'a-pat1))
	(new-var2 (make-variable-term *cosmos* 'a-pat2)))
    (list (make-right-assoc-normal-form top
					(cons new-var1
					      (list-assoc-subterms
					       term
					       (term-head term)
					       )))
	  (make-right-assoc-normal-form top
					(append
					 (list-assoc-subterms
					  term
					  (term-head term))
					 (list new-var1)))
	  (make-right-assoc-normal-form top
					(list new-var2
					      term
					      new-var1)))))
					 
(defun @pat-match (pat term)
  (multiple-value-bind (gs sub no-match eeq)
      (@matcher pat term)
    (unless no-match
      (return-from @pat-match (values t sub)))
    (if (and (term-is-application-form? term)
	     (term-is-application-form? pat)
	     (method-is-of-same-operator (term-head pat)
					 (term-head term)))
	(let ((top (term-head pat)))
	  (if (method-is-associative top)
	      (dolist (npat (if (method-is-commutative top)
				(list (@make-ac-pattern top pat))
				(@make-a-patterns top pat))
		       (values nil nil))
		(when (and npat
			   (progn
			     (multiple-value-setq (gs sub no-match eeq)
			       (@matcher npat term))
			     (null no-match)))
		  (return (values t sub))))
	      (values nil nil))))))

(defun @test-rule-extensions (rule term)
  (let ((top (term-head (axiom-lhs rule))))
    (if (method-is-associative top)
	(dolist (r (if (method-is-commutative top)
		       (compute-AC-extension rule top)
		       (compute-A-extensions rule top))
		 nil)
	  (when (and r (@test-rule-direct r term))
	    (return t)))
	nil)))

#|| moved to construct/axiom.lisp

;;; ***********************
;;; FINDING SPECIFIED RULES
;;; ***********************

;;; get-rule-numbered : mdoule number -> rule
;;;
(defun get-rule-numbered (mod n)
  (!setup-reduction mod)
  (when (<= n 0)
    (with-output-chaos-error ('invalid-rule-number)
      (format t "rule number must be greater than 0.")
      ))
  (let* ((*module-all-rules-every* t)
	 (res (nth (1- n) (get-module-axioms mod t))))
      (if (null res)
	  (with-output-chaos-error ('selection-out-of-bound)
	    (format t "selected rule doesn't exist, ~d" n)
	    )
	  res)))

(defun get-all-rules-labelled (mod l)
  (!setup-reduction mod)
  (let ((res nil)
	(*module-all-rules-every* t))
    (dolist (rul (get-module-axioms mod t))
      (when (memq (intern l) (axiom-labels rul))
	(push rul res)))
    res))

(defun get-rule-labelled (mod l)
  (let ((val (get-all-rules-labelled mod l)))
    (if (null val)
	(with-output-chaos-error ('no-rule-label)
	  (format t "No rule with label: ~a" l)
	  )
	(if (and val (null (cdr val)))
	    (car val)
	    (progn
	      (princ "no unique rule with label: ") (princ l) (terpri)
	      (chaos-error 'ambiguous-label)))
	)))

||#

;;; *********************
;;; VARIABLE SUBSTITUTION
;;; *********************

;;; COMPUTE-VARIABLE-SUBSTITUTION
;;;
(defun compute-variable-substitution (rule substtoks)
  ;; rule just for vars
  (let ((vars (union (term-variables (axiom-lhs rule))
		     (union (term-variables (axiom-rhs rule))
			    (term-variables (axiom-condition rule)))))
	(sub nil)
	varnm
	trmtoks
	avar
	aterm)
    ;; (!setup-parse *last-module*)
    (with-in-module (*last-module*)
      (loop (when (null substtoks) (return))
	    ;; <varid> = <term>
	    (setq varnm (cadr substtoks))
	    (setq trmtoks (nth 3 substtoks))
	    (setq avar (find-if #'(lambda (x) (equal (string (variable-name x))
						     varnm))
				vars))
	    (setq aterm (simple-parse *current-module*
				     trmtoks
				     *cosmos*))
	    (if (and avar (not (term-is-an-error aterm)))
		(progn
		  (if (not (is-in-same-connected-component
			    (term-sort aterm)
			    (variable-sort avar)
			    *current-sort-order*))
		      (with-output-chaos-warning ()
			(princ "sort of term is incompatible with variable sort")
			(print-next)
			(format t "variable ~a:" (string (variable-name avar)))
			(print-sort-name (variable-sort avar))
			(print-next)
			(princ "term ")
			(print-chaos-object aterm)
			(princ ":")
			(print-sort-name (term-sort aterm)))
		      (push (cons avar aterm) sub)))
		(with-output-chaos-error ('invalid-subst)
		  (unless avar
		    (format t "No such variable in rule: ~s" varnm)
		    (print-next)
		    (princ "specified substitution contains an error")
		    )))
	    (setq substtoks (cddddr substtoks))))
    sub))

;;; **********************
;;; FINDING MATCHING RULES
;;; **********************

(defstruct found-pattern
  rule-num
  direction
  rule
  subst
  extra)

(defun find-matching-rules (what target module)
  (with-in-module (module)
    (let* ((*module-all-rules-every* t)
	   (rules (get-module-axioms *current-module* t))
	   (res nil))
      (do* ((rls rules (cdr rls))
	    (rule (car rls) (car rls))
	    (num 1 (1+ num)))
	   ((endp rls))
	(when (or (eq what :rule) (eq what :+rule))
	  (multiple-value-bind (match subst)
	      (@pat-match (axiom-lhs rule) target)
	    (when match
	      (push (make-found-pattern :rule-num num
					:direction :+rule
					:rule rule
					:subst subst
					:extra (compute-extra-variables
						rule
						:+rule))
		    res))))
	(when (and (or (eq what :rule) (eq what :-rule))
		   (not (eq (axiom-type rule) :rule))
		   (not (rule-is-builtin rule)))
	  (multiple-value-bind (match subst)
	      (@pat-match (axiom-rhs rule) target)
	    (when match
	      (push (make-found-pattern :rule-num num
					:direction :-rule
					:rule rule
					:subst subst
					:extra (compute-extra-variables
						rule
						:-rule))
		    res)))))
      (nreverse res))))

;;; not used
#||
(defun make-variable-psuedo-constant (var)
  (make-pvariable-term (variable-sort var) (variable-name var)))
||#

(defun compute-extra-variables (rule direction)
  (let ((lhs (axiom-lhs rule))
	(rhs (axiom-rhs rule))
	(condition (axiom-condition rule)))
    (when (eq direction ':-rule)
      (setq lhs rhs)
      (setq rhs lhs))
    (let* ((lvars (term-variables lhs))
	   (rvars (union (term-variables rhs)
			 (term-variables condition))))
      (nset-difference rvars lvars))))

;;; EOF
