;;;-*-Mode:LISP; Package: CHAOS; Base:10; Syntax:Common-lisp -*-
;;; $Id: cexec.lisp,v 1.6 1998/07/16 08:19:06 sawada Exp $
(in-package :chaos)
#|=============================================================================
			     System:CHAOS
			    Module:thstuff
			   File:cexec.lisp
=============================================================================|#
#-:chaos-debug
(declaim (optimize (speed 3) (safety 0) #-GCL (debug 0)))
#+:chaos-debug
(declaim (optimize (speed 1) (safety 3) #-GCL (debug 3)))

;;;
;;;
;;;
(defun make-rule-pat (pos rule)
  (list pos rule))

(defmacro rule-pat-pos (pat) `(car ,pat))
(defmacro rule-pat-rule (pat) `(cadr ,pat))

(defun find-matching-rules-for-exec (target module &optional start-pos)
  (when start-pos
    (setq target (get-target-subterm target start-pos)))
  ;;
  (with-in-module (module)
    (let* ((*module-all-rules-every* t)
	   (rules (get-module-axioms *current-module* t))
	   (rls nil)
	   (res nil))
      (dolist (rule rules)
	(when (rule-is-rule rule) (push rule rls)))
      (setq res (find-matching-rules-for-exec* target rls start-pos))
      (delete-duplicates res
			 :test #'(lambda (x y)
				   (and (equal (rule-pat-pos x)
					       (rule-pat-pos y))
					(eq (rule-pat-rule x)
					    (rule-pat-rule y))))))))

(defun find-matching-rules-for-exec* (target rules pos)
  (if (term-is-application-form? target)
      (let ((res nil))
	(do* ((rls rules (cdr rls))
	      (rule (car rls) (car rls)))
	     ((endp rls))
	  (let ((patterns nil)
		(head (term-head (rule-lhs rule))))
	    (push rule patterns)
	    (when (method-is-associative head)
	      (if (method-is-commutative head)
		  (let ((ac-ext (!axiom-ac-extension rule)))
		    (unless ac-ext 
		      (setq ac-ext (compute-AC-extension rule head)))
		    (push (car ac-ext) patterns))
		  ;;
		  (let ((a-exts (!axiom-A-extensions rule)))
		    (unless a-exts
		      (setq a-exts
			    (compute-A-extensions rule head)))
		    (dolist (r a-exts)
		      (push r patterns)))))
	    ;;
	    (dolist (pat patterns)
	      ;; (break)
	      (when (cexec-pat-match (axiom-lhs pat) target)
		(push (make-rule-pat pos pat) res)))
	    ))
	;;
	(dotimes (x (length (term-subterms target)))
	  (let ((r (find-matching-rules-for-exec*
		    (term-arg-n target x)
		    rules
		    (append pos (list x)))))
	    (when r (setq res (nconc res r)))))
	res)
      nil))

(defun cexec-pat-match (pat target)
  (multiple-value-bind (gs sub no eeq)
      (@matcher pat target)
    (declare (ignore gs sub eeq))
    (null no)))

(defun get-target-subterm (term pos)
  (let ((cur term))
    (when pos
      (dolist (p pos)
	(setq cur (term-arg-n cur p))
	(unless cur (with-output-panic-message ()
		      (format t "could not find subterm at pos ~d" p)
		      (format t "~% target was ")
		      (term-print cur)
		      (chaos-error 'panic)))
	))
    cur))

;;;
;;;
;;;
;;; (defvar *cexec-trace* nil)
(defvar *cexec-current-step* 1)

(defun mark-term-as-not-reduced-full (term)
  (when (term-is-builtin-constant? term)
    (return-from mark-term-as-not-reduced-full nil))
  (mark-term-as-not-reduced term)
  (when (term-is-application-form? term)
    (dolist (sub (term-subterms term))
      (mark-term-as-not-reduced-full sub))))

(defun make-cterm-node (num term)
  (list num term nil))

(defmacro cterm-node-pos (cterm-node)
  `(car ,cterm-node))
(defmacro cterm-node-term (cterm-node)
  `(cadr ,cterm-node))
(defmacro cterm-node-children (cterm-node)
  `(caddr ,cterm-node))

;;; HASH
(defvar *cexec-term-hash* nil)
(defun clear-cexec-hash ()
  (setq *cexec-term-hash* nil))

(defun set-cexec-hash (term)
  (push term *cexec-term-hash*))

(defun find-in-cexec-hash (term)
  (find-if #'(lambda (x) (term-equational-equal term x))
	   *cexec-term-hash*))

#||
(defconstant .cexec-term-hash-size. 1001)
(eval-when (eval load)
  (setq *cexec-term-hash* (alloc-svec .cexec-term-hash-size.)))

(defun clear-cexec-hash ()
  (dotimes (x .cexec-term-hash-size.)
    (setf (svref *cexec-term-hash* x) nil))
  *cexec-term-hash*)

#-GCL
(declaim (inline find-in-cexec-hash))

(#-GCL defun #+GCL si:define-inline-function
       find-in-cexec-hash (term)
  (let ((val (hash-term term)))
    (declare (type term-hash-key val))
    (let* ((ent (svref *cexec-term-hash*
		       (mod val .cexec-term-hash-size.)))
	   (val (cdr (assoc term ent :test #'term-is-similar?))))
      val)))

#-GCL
(declaim (inline set-cexec-hash))

(#-GCL defun #+GCL si:define-inline-function
       set-cexec-hash (term)
  (let ((val (hash-term term)))
    (declare (type term-hash-key val))
    (let ((ind (mod val .cexec-term-hash-size.)))
      (let ((ent (svref *cexec-term-hash* ind)))
	(let ((pr (assoc term ent :test #'term-is-similar?)))
	  (if pr (rplacd pr term)
	      (setf (svref *cexec-term-hash* ind)
		    (cons (cons term term) ent)))
	  )))))

||#
;;;
(defvar *cexec-debug* nil)

(defun cexec-term-1 (cterm-node &optional
				subtimes
				(module (or *current-module*
					    *last-module*))
				 &aux (term (cterm-node-term cterm-node)))
  ;;
  (with-in-module (module)
    (let ((xterm (simple-copy-term term))
	  (ptrans? nil))
      ;;
      (if (find-in-cexec-hash xterm)
	  (progn
	    (when *cexec-trace*
	      (with-output-msg ()
		(princ "loop in transitions via term : ")
		(let ((*print-indent* (+ *print-indent* 4)))
		  (print-chaos-object xterm))))
	    (return-from cexec-term-1 nil))
	  (set-cexec-hash (simple-copy-term xterm)))
      ;;
      (when *cexec-debug*
	(format t "~% target = ")
	(let ((*fancy-print* nil))
	  (term-print xterm))
	(flush-all))
      ;;
      (let ((*rewrite-exec-mode* nil)
	    ;; (*closed-world* nil)
	    )
	(when *cexec-normalize*
	  (rewrite* xterm)))
      ;;
      (when *cexec-trace*
	(print-next)
	(unless (term-equational-equal xterm term)
	  (flush-all)
	  (let ((*fancy-print* nil)
		(*print-indent* (+ 4 *print-indent*)))
	    (princ "> pre normalization : ")
	    (term-print term)
	    (print-next)
	    (princ "== ")
	    (term-print xterm))
	  (flush-all)))
      ;;
      (let ((rule-pats (find-matching-rules-for-exec xterm module))
	    (*rewrite-exec-mode* t)
	    (res nil))
	;;
	(when *cexec-debug*
	  (dolist (rpat rule-pats)
	    (format t "~&~a: "(car rpat))
	    (print-chaos-object (cadr rpat))))
	;;
	(unless rule-pats
	  (when *cexec-trace*
	    (when (and (term-is-applform? term)
		       (method-has-trans-rule (term-head term)))
	      (with-output-simple-msg ()
		(format t "-- no more possible transition for : ")
		(term-print term))))
	  (return-from cexec-term-1 nil))
	;;
	(when *cexec-trace*
	  (flush-all)
	  (if (not subtimes)
	      (format t "~%** transition step ~D **********" *cexec-current-step*)
	      (format t "~%** transition step ~D-~D *******"
		      *cexec-current-step*
		      subtimes))
	  (flush-all))
	;;
	(do* ((rls rule-pats (cdr rls))
	      (rule (car rls) (car rls)))
	     ((endp rls))
	  (let* ((target-whole (simple-copy-term xterm))
		 (target (get-target-subterm target-whole
					     (rule-pat-pos rule))))
	    ;; check hash
	    ;; the following should be done iff the target is NOT
	    ;; in hash table + register target in hash.
	    ;;
	    (when (apply-rule (rule-pat-rule rule) target)
		(when *cexec-debug*
		  (format t "~%returing : ")
		  (print-chaos-object target)
		  (flush-all))
		(when *cexec-trace*
		  (print-next)
		  (flush-all)
		  (let ((*fancy-print* nil)
			(*print-indent* (+ 4 *print-indent*)))
		    (format t "(~{~d~^ ~}): " (mapcar #'1+ (rule-pat-pos rule)))
		    (term-print xterm)
		    (print-next)
		    (format t "=(~D)=> " *cexec-current-step*)
		    (term-print target-whole))
		  (flush-all))
		(push (list (rule-pat-pos rule) target-whole) res))
	    ))
	;;
	(let ((*rewrite-exec-mode* nil)
	      ;; (*closed-world* nil)
	      (org nil))
	  (dolist (x res)
	    (when *cexec-trace*
	      (setq org (simple-copy-term (cadr x))))
	    ;;
	    (mark-term-as-not-reduced-full (cadr x))
	    (when *cexec-normalize*
	      (rewrite* (cadr x)))
	    ;;
	    (when *cexec-trace*
	      (let ((*fancy-print* nil))
		(unless (term-equational-equal org (cadr x))
		  (print-next)
		  (flush-all)
		  (let ((*print-indent* (+ 4 *print-indent*)))
		    (princ "< post nomalization : ")
		    (term-print org)
		    (print-next)
		    (princ "== ")
		    (term-print (cadr x)))
		  (flush-all))))
	    (unless (member (cadr x) (cterm-node-children cterm-node)
			    :test #'(lambda (t1 pat)
				      (term-equational-equal t1
							     (cadr pat))))
	      (push (nconc x (list nil))
		    (cterm-node-children cterm-node))
	      (setq ptrans? t))))
	(if ptrans?
	    cterm-node
	    nil)))))

(defun term-pattern-included-in-cexec (t1 t2 max
					  &optional (module
						     (or *current-module*
							 *last-module*)))
  (setq *cexec-current-step* 1)
  (with-in-module (module)
    (let ((num (if (term-is-builtin-constant? max)
		   (term-builtin-value max)
		   *cexec-limit*))
	  (*rewrite-semantic-reduce* (module-has-behavioural-axioms module))
	  (result nil))
      (declare (special *rewrite-semantic-reduce*))
      ;;
      (when *cexec-normalize*
	(rewrite* t2))
      (when (term-equational-equal t1 t2)
	(return-from term-pattern-included-in-cexec t))
      ;; 
      ;; clear hash 
      (clear-cexec-hash)
      ;;
      ;; real work
      (setq result
	    (term-pattern-included-in-cexec* (list (make-cterm-node nil t1))
					     t2
					     num))
      ;; 
      #||
      (when *cexec-target*
	(when *cexec-debug*
	  (format t "~%marking reduced: ")
	  (print-chaos-object *cexec-target*))
	(mark-term-as-reduced *cexec-target*))
      ||#
      ;;
      (if (eq result :limit)
	  (with-output-chaos-warning ()
	    (format t
		    "=(N)=> terminates, reaching preset limit of transitions."))
	  result)
      )))

(defun term-pattern-included-in-cexec* (t-nodes t2 max)
  (block
      exit
    ;;
    (let ((no-trans nil)
	  (trans? nil)
	  (subtimes 0))
      (loop
       ;;
       (unless t-nodes (return-from exit nil))
       ;;
       (decf max)
       (when (< max 0)
	 (return-from exit :limit))
       ;;
       (setq subtimes 0)
       (dolist (t-node t-nodes)
	 (incf subtimes)
	 (setq trans? (cexec-term-1 t-node subtimes))
	 (if trans?
	     (let ((nexts (cterm-node-children t-node)))
	       (dolist (nx nexts)
		 (multiple-value-bind (match subst)
		     (@pat-match t2 (cterm-node-term nx))
		   (when match
		     (when *cexec-trace*
		       (print-next)
		       (flush-all)
		       (princ "** term : ")
		       (let ((*print-indent* (+ 4 *print-indent*)))
			 (term-print (cterm-node-term nx)))
		       (print-next)
		       (princ " matched to the pattern : ")
		       (let ((*print-indent* (+ 4 *print-indent*)))
			 (term-print t2))
		       (print-next)
		       (princ " with the substitution : ")
		       (let ((*print-indent* (+ 4 *print-indent*)))
			 (print-substitution subst))
		       (flush-all))
		     (return-from exit (list match subst))))))
	     (push t-node no-trans)))
       ;;
       (incf *cexec-current-step*)
       ;;
       (dolist (x no-trans)
	 (setq t-nodes (delete x t-nodes)))
       ;;
       (setq t-nodes
	     (mapcan #'(lambda (x)
			 (cterm-node-children x))
		     t-nodes))
       ))))

;;; EOF
