;;;-*- Mode:LISP; Package:CHAOS; Base:10; Syntax:Common-lisp -*-
;;; $Id: eval-match.lisp,v 1.1.1.1 1998/05/16 11:27:50 sawada Exp $
(in-package :chaos)
#|==============================================================================
				 System: CHAOS
				Module: thstuff
			     File: eval-match.lisp
==============================================================================|#
#-:chaos-debug
(declaim (optimize (speed 3) (safety 0) #-GCL (debug 0)))
#+:chaos-debug
(declaim (optimize (speed 1) (safety 3) #-GCL (debug 3)))

;;; DESCRIPTION ----------------------------------------------------------------
;;; evaluators of match, find commands.
;;; *NOTE* find <what> is just the alias of match it to <what> .
;;;

;;; ******
;;; MATCH
;;; ******

(defun eval-match-command (ast)
  (unless *last-module*
    (with-output-chaos-error ('no-current-module)
      (princ "no current module.")
      ))
  (let ((target (case (%match-target ast)
		  (:top $$term)
		  (:subterm $$subterm)
		  (:it (if $$subterm
			   $$subterm
			   $$term))
		  (t (let* ((*parse-variables* nil)
			    (parsed (with-in-module (*last-module*)
				      (simple-parse *current-module*
						    (%match-target ast)
						    *cosmos*))))
		       (if (sort<= (term-sort parsed)
				   *syntax-err-sort*
				   *chaos-sort-order*)
			   nil
			   parsed))))))
    ;;
    (when (or (null target) (eq target 'void))
      (if (symbolp (%match-target ast))
	  (with-output-chaos-warning ()
	    (princ "no target term is specified, please `start' or `choose'.")
	    ))
      (return-from eval-match-command nil))
    ;;
    (if (symbolp (%match-pattern ast))
	(find-rewrite-rules target (%match-pattern ast))
	(perform-match target (%match-pattern ast)))))

(defun find-rewrite-rules (target what)
  (let* ((real-target (supply-psuedo-variables target))
	 (patterns (find-matching-rules what real-target *last-module*)))
    (unless patterns
      (with-in-module (*last-module*)
	(format t "~&no rules found for term : ")
	(term-print target))
      (return-from find-rewrite-rules nil))
    ;; report the result
    (format t "~&== matching rules to term : ")
    (with-in-module (*last-module*)
      (let ((*fancy-print* nil))
	(term-print target))
      (dolist (pat patterns)
	(let ((num (found-pattern-rule-num pat))
	      (direction (found-pattern-direction pat))
	      (rule (found-pattern-rule pat))
	      (subst (found-pattern-subst pat))
	      (extra (found-pattern-extra pat)))
	  (print-next)
	  (if (eq direction :-rule)
	      (princ "-")
	      (princ "+"))
	  (print-mod-name *current-module*)
	  (if (rule-labels rule)
	      (format t ".{ ~D " num)
	      (format t ".~D" num))
	  (dolist (label (rule-labels rule))
	    (print-check)
	    (princ "| ")
	    (princ (string label)))
	  (when (rule-labels rule)
	    (princ " }"))
	  (princ " is ")
	  (print-axiom-brief rule nil t)
	  (format t "~& substitution : ")
	  (let ((*print-indent* (+ *print-indent* 4)))
	    (print-substitution subst))
	  (when extra
	    (format t "~& extra variables : ")
	    (format t "~{~a~^ ~}" (mapcar #'(lambda (x) (string (variable-name x)))
					  extra)))
	  )))
    ))

(defvar *use-choose-match* nil)

(defun perform-match (target pre-pattern)
  (let ((real-target (supply-psuedo-variables target)))
    (let ((first-match-meth (if *use-choose-match*
				nil
				'@matcher))
	  (next-match-meth (if *use-choose-match*
			       nil
			       'next-match)))
      (with-in-module (*last-module*)
	(let* ((*parse-variables* nil)
	       (pattern (simple-parse *current-module*
					pre-pattern
					*cosmos*)))
	  (when (sort<= (term-sort pattern)
			*syntax-err-sort* *chaos-sort-order*)
	    (return-from perform-match nil))
	  ;; normal case.
	  #||
	  (princ "-- match")
	  (princ " : ")
	  (term-print real-target)
	  (print-next)
	  (princ "   to    : ")
	  (term-print pattern)
	  (print-next)
	  ||#
	  ;;
	  (when *use-choose-match*
	    (let ((meth (choose-match-method real-target *bool-true* nil)))
	      (setf first-match-meth (car meth))
	      (setf next-match-meth (cdr meth))))
	  ;; ---- first match 
	  (multiple-value-bind (global-state subst no-match e-equal)
	      (funcall first-match-meth pattern real-target)
	    (when no-match
	      (format t "~&-- no match")
	      (return-from perform-match nil))
	    (format t "~&-- match success.")
	    (when e-equal
	      (format t "~&-- given terms are equational equal.")
	      (return-from perform-match nil))
	    (format t "~& substitution : ")
	    (let ((*print-indent* (+ *print-indent* 4)))
	      (print-substitution subst))
	    ;; ---- next matches
	    (block end
	      (multiple-value-setq (global-state subst no-match)
		(funcall next-match-meth global-state))
	      (while (not no-match)
		(cond ((y-or-n-p-wait #\y 20 ">> More? [y/n] : ")
		       (format t "~&-- match success : ")
		       (let ((*print-indent* (+ 4 *print-indent*)))
			 (format t "~& * substitution : ")
			 (print-substitution subst))
		       (print-next))
		      (t (return-from end)))
		(multiple-value-setq (global-state subst no-match)
		  (funcall next-match-meth global-state)))
	      (format t "~%-- No more match")
	      )))))))

;;; EOF
